@@ -26,7 +26,7 @@ sig
   val hash : t -> int
 end
 
-module Hashcons (M : OrderedType) (H : HashedType with type t = M.t) : Hashcons.S with
+module Hashcons (M : OrderedType) (_ : HashedType with type t = M.t) : Hashcons.S with
   type t = Set.Make(M).t
   and type u = M.t -> M.t
 (** Create hash-consing for sets. The hashing function provided must be
