@@ -5,10 +5,11 @@ set -e
 ci_dir="$(dirname "$0")"
 . "${ci_dir}/ci-common.sh"
 
-FORCE_GIT=1
+WITH_SUBMODULES=1
 git_download cross_crypto
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 ( cd "${CI_BUILD_DIR}/cross_crypto"
-  git submodule update --init --recursive
   make
 )
