@@ -7,12 +7,17 @@ ci_dir="$(dirname "$0")"
 
 git_download menhirlib
 
-( cd "${CI_BUILD_DIR}/menhirlib"
-  dune build @install -p menhirLib,menhirSdk,menhir
-  dune install -p menhirLib,menhirSdk,menhir menhir menhirSdk menhirLib --prefix="$CI_INSTALL_DIR"
-)
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
 
 ( cd "${CI_BUILD_DIR}/menhirlib"
+  if grep -q unreleased dune-project; then
+    date=$(date +%Y%m%d)
+    sed -i.bak "s/unreleased/$date/" dune-project
+    echo "Definition require_$date := tt." > coq-menhirlib/src/Version.v
+  fi
+  dune build @install -p menhirLib,menhirSdk,menhir
+  dune install -p menhirLib,menhirSdk,menhir menhir menhirSdk menhirLib --prefix="$CI_INSTALL_DIR"
+
   make -C coq-menhirlib
   make -C coq-menhirlib install
 )
