@@ -7,7 +7,9 @@ ci_dir="$(dirname "$0")"
 
 git_download quickchick
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 ( cd "${CI_BUILD_DIR}/quickchick"
-  make
-  make install-plugin
+  dune build -p coq-quickchick @install
+  dune install -p coq-quickchick --prefix=$CI_INSTALL_DIR
 )
