@@ -7,7 +7,8 @@ ci_dir="$(dirname "$0")"
 
 git_download rewriter
 
-export COQEXTRAFLAGS='-native-compiler no'
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 ( cd "${CI_BUILD_DIR}/rewriter"
   make
   make install
