@@ -52,7 +52,15 @@ do
     _CONTRIB_NAME=$1
     _CONTRIB_GITURL=${_CONTRIB_NAME}_CI_GITURL
     _CONTRIB_GITURL=${!_CONTRIB_GITURL}
+    _CONTRIB_SUBMODULE_GITURL=${_CONTRIB_NAME}_CI_SUBMODULE_GITURL
+    _CONTRIB_SUBMODULE_GITURL=${!_CONTRIB_SUBMODULE_GITURL}
+    _CONTRIB_SUBMODULE_BRANCH=${_CONTRIB_NAME}_CI_SUBMODULE_BRANCH
+    _CONTRIB_SUBMODULE_BRANCH=${!_CONTRIB_SUBMODULE_BRANCH}
+    if [[ -n "${_CONTRIB_SUBMODULE_GITURL}" ]]; then
+        _CONTRIB_GITURL="${_CONTRIB_SUBMODULE_GITURL}"
+    fi
     echo "Processing Contrib $_CONTRIB_NAME"
+    shift
 
     # check _CONTRIB_GIT exists and it is of the from github...
 
@@ -72,8 +80,10 @@ do
     setup_contrib_git $_CONTRIB_DIR $_CONTRIB_GITPUSHURL
 
     echo "overlay ${_CONTRIB_NAME} $_CONTRIB_GITURL $OVERLAY_BRANCH $PR_NUMBER" >> $OVERLAY_FILE
-    echo "" >> $OVERLAY_FILE
-    shift
+    if [ -n "${_CONTRIB_SUBMODULE_BRANCH}${_CONTRIB_SUBMODULE_GITURL}" ]; then
+        echo "# Make PRs against ${_CONTRIB_SUBMODULE_GITURL} base branch ${_CONTRIB_SUBMODULE_BRANCH}" >> $OVERLAY_FILE
+    fi
+    if [ $# -gt 0 ]; then echo "" >> $OVERLAY_FILE; fi
 done
 
 # Copy to overlays folder.
