@@ -12,10 +12,11 @@ open Names
 open Constr
 open Environ
 open Univ
+open UVars
 
 type univ_length_mismatch = {
-  actual : int ;
-  expect : int ;
+  actual : int * int;
+  expect : int * int;
 }
 (* Due to an OCaml bug ocaml/ocaml#10027 inlining this record will cause
 compliation with -rectypes to crash. *)
@@ -32,30 +33,47 @@ val new_global_univ : unit -> Universe.t in_universe_context_set
 (** Build a fresh instance for a given context, its associated substitution and
     the instantiated constraints. *)
 
-val fresh_instance : AbstractContext.t -> Instance.t in_universe_context_set
+type sort_context_set = (Sorts.QVar.Set.t * Univ.Level.Set.t) * Univ.Constraints.t
+
+type 'a in_sort_context_set = 'a * sort_context_set
+
+val sort_context_union : sort_context_set -> sort_context_set -> sort_context_set
+
+val empty_sort_context : sort_context_set
+
+val is_empty_sort_context : sort_context_set -> bool
+
+val diff_sort_context : sort_context_set -> sort_context_set -> sort_context_set
+
+val fresh_instance : AbstractContext.t -> Instance.t in_sort_context_set
 
 val fresh_instance_from : ?loc:Loc.t -> AbstractContext.t -> Instance.t option ->
-  Instance.t in_universe_context_set
+  Instance.t in_sort_context_set
 
 val fresh_sort_in_family : Sorts.family ->
-  Sorts.t in_universe_context_set
-val fresh_constant_instance : env -> Constant.t ->
-  pconstant in_universe_context_set
-val fresh_inductive_instance : env -> inductive ->
-  pinductive in_universe_context_set
-val fresh_constructor_instance : env -> constructor ->
-  pconstructor in_universe_context_set
-val fresh_array_instance : env ->
-  Instance.t in_universe_context_set
+  Sorts.t in_sort_context_set
+(** NB: InQSort is treated as InType *)
 
-val fresh_global_instance : ?loc:Loc.t -> ?names:Univ.Instance.t -> env -> GlobRef.t ->
-  constr in_universe_context_set
+val fresh_constant_instance : env -> Constant.t ->
+  pconstant in_sort_context_set
+val fresh_inductive_instance : env -> inductive ->
+  pinductive in_sort_context_set
+val fresh_constructor_instance : env -> constructor ->
+  pconstructor in_sort_context_set
+val fresh_array_instance : env ->
+  Instance.t in_sort_context_set
+
+val fresh_global_instance : ?loc:Loc.t -> ?names:UVars.Instance.t -> env -> GlobRef.t ->
+  constr in_sort_context_set
 
 (** Get fresh variables for the universe context.
     Useful to make tactics that manipulate constrs in universe contexts polymorphic. *)
 val fresh_universe_context_set_instance : ContextSet.t ->
   universe_level_subst * ContextSet.t
 
+val fresh_sort_context_instance : sort_context_set ->
+  sort_level_subst * sort_context_set
+
 (** Create a fresh global in the environment argument, without side effects.
     BEWARE: this raises an error on polymorphic constants/inductives:
     the constraints should be properly added to an evd.
