@@ -66,12 +66,13 @@ let coqide_known_option table = List.mem table [
   ["Printing";"Existential";"Instances"];
   ["Printing";"Universes"];
   ["Printing";"Unfocused"];
+  ["Printing";"Goal";"Names"];
   ["Diffs"]]
 
 let is_known_option cmd = match cmd with
-  | VernacSetOption (_, o, OptionSetTrue)
-  | VernacSetOption (_, o, OptionSetString _)
-  | VernacSetOption (_, o, OptionUnset) -> coqide_known_option o
+  VernacSynterp (VernacSetOption (_, o, OptionSetTrue)
+    | VernacSetOption (_, o, OptionSetString _)
+    | VernacSetOption (_, o, OptionUnset)) -> coqide_known_option o
   | _ -> false
 
 let ide_cmd_warns ~id { CAst.loc; v } =
@@ -87,21 +88,19 @@ let ide_doc = ref None
 let get_doc () = Option.get !ide_doc
 let set_doc doc = ide_doc := Some doc
 
-(* This patches up the offset to compensate the XML protocol not
-   having a notion of absolute offset *)
-let fix_loc ~off loc =
-  Loc.{ loc with bol_pos = loc.bol_pos + off;
-                 bp = loc.bp + off;
-                 ep = loc.ep + off}
-
 let add ((((s,eid),(sid,verbose)),off),(line_nb,bol_pos)) =
   let doc = get_doc () in
   let open Loc in
-  (* note: this won't yield correct values for bol_pos_last,
-     but the debugger doesn't use that *)
-  let loc = { (initial ToplevelInput) with bp=off; line_nb } in
-  let fix_loc = fix_loc ~off in
-  let pa = Pcoq.Parsable.make ~loc ~fix_loc (Gramlib.Stream.of_string s) in
+  (* This needs to be akin to what CLexer.after does *)
+  let loc = { (initial ToplevelInput) with
+              bp=off
+            ; line_nb
+            ; bol_pos
+            ; ep=off
+            ; line_nb_last=line_nb
+            ; bol_pos_last=bol_pos } in
+  let r_stream = Gramlib.Stream.of_string ~offset:off s in
+  let pa = Pcoq.Parsable.make ~loc r_stream in
   match Stm.parse_sentence ~doc sid ~entry:Pvernac.main_entry pa with
   | None -> assert false (* s may not be empty *)
   | Some ast ->
@@ -193,7 +192,7 @@ let concl_next_tac =
   ])
 
 let process_goal short sigma g =
-  let evi = Evd.find sigma g in
+  let evi = Evd.find_undefined sigma g in
   let env = Evd.evar_filtered_env (Global.env ()) evi in
   let min_env = Environ.reset_context env in
   let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name env sigma g)) else None in
@@ -202,13 +201,13 @@ let process_goal short sigma g =
   in
   let process_hyp d (env,l) =
     let d' = CompactedDecl.to_named_context d in
-      (List.fold_right Environ.push_named d' env,
-       (pr_compacted_decl env sigma d) :: l) in
+      (List.fold_right EConstr.push_named d' env,
+       (pr_ecompacted_decl env sigma d) :: l) in
   let hyps =
     if short then [] else
       let (_env, hyps) =
         Context.Compacted.fold process_hyp
-        (Termops.compact_named_context (Environ.named_context env)) ~init:(min_env,[])
+        (Termops.compact_named_context sigma (EConstr.named_context env)) ~init:(min_env,[])
       in
       hyps
   in
@@ -245,7 +244,7 @@ let export_pre_goals flags Proof.{ sigma; goals; stack } process =
 
 let subgoals flags =
   let doc = get_doc () in
-  set_doc @@ fst @@ Stm.finish ~doc;
+  ignore (Stm.finish ~doc : Vernacstate.t);
   let short = match flags.Interface.gf_mode with
   | "short" -> true
   | _ -> false
@@ -276,7 +275,7 @@ let goals () =
 let evars () =
   try
     let doc = get_doc () in
-    set_doc @@ fst @@ Stm.finish ~doc;
+    ignore (Stm.finish ~doc : Vernacstate.t);
     let pfts = Vernacstate.Declare.give_me_the_proof () in
     let Proof.{ sigma } = Proof.data pfts in
     let exl = Evar.Map.bindings (Evd.undefined_map sigma) in
@@ -293,7 +292,7 @@ let hints () =
     match goals with
     | [] -> None
     | g :: _ ->
-      let env = Evd.evar_filtered_env (Global.env ()) (Evd.find sigma g) in
+      let env = Evd.evar_filtered_env (Global.env ()) (Evd.find_undefined sigma g) in
       let get_hint_hyp env d accu = hyp_next_tac sigma env d :: accu in
       let hint_hyps = List.rev (Environ.fold_named_context get_hint_hyp env ~init: []) in
       Some (hint_hyps, concl_next_tac)
@@ -304,15 +303,14 @@ let hints () =
 
 let wait () =
   let doc = get_doc () in
-  set_doc (Stm.wait ~doc)
+  Stm.wait ~doc
 
 let status force =
   (* We remove the initial part of the current [DirPath.t]
      (usually Top in an interactive session, cf "coqtop -top"),
      and display the other parts (opened sections and modules) *)
-  set_doc (fst @@ Stm.finish ~doc:(get_doc ()));
-  if force then
-    set_doc (fst @@ Stm.join ~doc:(get_doc ()));
+  ignore (Stm.finish ~doc:(get_doc ()) : Vernacstate.t);
+  if force then Stm.join ~doc:(get_doc ());
   let path =
     let l = Names.DirPath.repr (Lib.cwd ()) in
     List.rev_map Names.Id.to_string l
@@ -394,12 +392,10 @@ let import_option_value = function
 
 let export_option_state s = {
   Interface.opt_sync  = true;
-  Interface.opt_depr  = s.Goptions.opt_depr;
+  Interface.opt_depr  = Option.has_some s.Goptions.opt_depr;
   Interface.opt_value = export_option_value s.Goptions.opt_value;
 }
 
-exception NotSupported of string
-
 let proof_diff (diff_opt, sid) =
   let diff_opt = Proof_diffs.string_to_diffs diff_opt in
   let doc = get_doc () in
@@ -462,15 +458,12 @@ let about () = {
 
 let handle_exn (e, info) =
   let dummy = Stateid.dummy in
-  let loc_of e = match Loc.get_loc e with
-    | Some loc -> Some (Loc.unloc loc)
-    | _        -> None in
   let mk_msg () = CErrors.iprint (e,info) in
   match e with
   | e ->
       match Stateid.get info with
-      | Some (valid, _) -> valid, loc_of info, mk_msg ()
-      | None -> dummy, loc_of info, mk_msg ()
+      | Some (valid, _) -> valid, Loc.get_loc info, mk_msg ()
+      | None -> dummy, Loc.get_loc info, mk_msg ()
 
 let init =
   let initialized = ref false in
@@ -486,7 +479,7 @@ let init =
            get_doc (), init_sid, `NewTip in
          if Filename.check_suffix file ".v" then
            Stm.set_compilation_hints file;
-         set_doc (fst @@ Stm.finish ~doc);
+         ignore (Stm.finish ~doc : Vernacstate.t);
          initial_id
    end
 
@@ -574,7 +567,8 @@ let slave_feeder fmt xml_oc msg =
 
 let msg_format = ref (fun () ->
     let width = Option.default 72 (Topfmt.get_margin ()) in
-    Xmlprotocol.Richpp { width; depth = max_int }
+    let depth = Option.default max_int (Topfmt.get_depth_boxes ()) in
+    Xmlprotocol.Richpp { width; depth }
   )
 
 (* The loop ignores the command line arguments as the current model delegates
