@@ -93,6 +93,11 @@ let insert_with_tags (buf : #GText.buffer_skel) mark rmark tags text =
 
 let nl_white_regex = Str.regexp "^\\( *\n *\\)"
 let diff_regex = Str.regexp "^diff."
+let escape_regex = Str.regexp "[^\n\x20-\xff]\\|\x7f"
+
+let escaped s =
+  let subst s = Printf.sprintf "\\x%02x" (Char.code (Str.matched_string s).[0]) in
+  Str.global_substitute escape_regex subst s
 
 let insert_xml ?(mark = `INSERT) ?(tags = []) (buf : #GText.buffer_skel) msg =
   let open Xml_datatype in
@@ -116,7 +121,7 @@ let insert_xml ?(mark = `INSERT) ?(tags = []) (buf : #GText.buffer_skel) msg =
       insert_with_tags buf mark rmark etags s
   in
   let rec insert tags = function
-  | PCData s -> insert_str tags s
+  | PCData s -> insert_str tags (escaped s)
   | Element (t, _, children) ->
     let (pfx, tname) = Pp.split_tag t in
     let is_diff = try let _ = Str.search_forward diff_regex tname 0 in true with Not_found -> false in
@@ -169,6 +174,19 @@ let ulen uni_ch =
   else if uni_ch < 0x10000 then 3
   else 4
 
+(* workaround for lablgtk bug up to version 3.1.4 *)
+(* replaces: buffer#get_iter_at_byte ~line index *)
+(* see https://github.com/garrigue/lablgtk/pull/181 *)
+let get_iter_at_byte buffer ~line index =
+  let iter = buffer#get_iter_at_byte ~line 0 in
+  let rec adjust iter cnt =
+    if cnt <= 0 then
+      iter
+    else
+      adjust (iter#forward_char) (cnt - (ulen iter#char))
+  in
+  adjust iter index
+
 (** convert UTF-8 byte offset (used by Coq) to unicode offset (used by GTK buffer) *)
 let byte_off_to_buffer_off buffer byte_off =
   let rec cvt iter cnt =
