@@ -24,6 +24,8 @@ exception ModuleInternalizationError of module_internalization_error
 
 type module_kind = Module | ModType | ModAny
 
+type module_struct_expr = (universe_decl_expr option * constr_expr) Declarations.module_alg_expr
+
 let error_not_a_module_loc ~info kind loc qid =
   let e = match kind with
     | Module -> NotAModule qid
@@ -79,11 +81,11 @@ let lookup_polymorphism env base kind fqid =
     | [id] ->
       let test (lab,obj) =
         match Id.equal (Label.to_id lab) id, obj with
-        | false, _ | _, (SFBmodule _ | SFBmodtype _) -> None
+        | false, _ | _, (SFBrules _ | SFBmodule _ | SFBmodtype _) -> None
         | true, SFBmind mind -> Some (Declareops.inductive_is_polymorphic mind)
         | true, SFBconst const -> Some (Declareops.constant_is_polymorphic const)
       in
-      (try CList.find_map test m with Not_found -> false (* error later *))
+      (match CList.find_map test m with Some v -> v | None -> false (* error later *))
     | id::rem ->
       let next = function
         | MoreFunctor _ -> false (* error later *)
@@ -91,12 +93,12 @@ let lookup_polymorphism env base kind fqid =
       in
       let test (lab,obj) =
         match Id.equal (Label.to_id lab) id, obj with
-        | false, _ | _, (SFBconst _ | SFBmind _) -> None
+        | false, _ | _, (SFBrules _ | SFBconst _ | SFBmind _) -> None
         | true, SFBmodule body -> Some (next body.mod_type)
         | true, SFBmodtype body ->  (* XXX is this valid? If not error later *)
           Some (next body.mod_type)
       in
-      (try CList.find_map test m with Not_found -> false (* error later *))
+      (match CList.find_map test m with Some v -> v | None -> false (* error later *))
   in
   aux (defunctor m) fqid
 
@@ -135,8 +137,8 @@ let interp_with_decl env base kind = function
     let poly = lookup_polymorphism env base kind fqid in
     begin match fst (UState.check_univ_decl ~poly ectx udecl) with
       | UState.Polymorphic_entry ctx ->
-        let inst, ctx = Univ.abstract_universes ctx in
-        let c = EConstr.Vars.subst_univs_level_constr (Univ.make_instance_subst inst) c in
+        let inst, ctx = UVars.abstract_universes ctx in
+        let c = EConstr.Vars.subst_univs_level_constr (UVars.make_instance_subst inst) c in
         let c = EConstr.to_constr sigma c in
         WithDef (fqid,(c, Some ctx)), Univ.ContextSet.empty
       | UState.Monomorphic_entry ctx ->
