@@ -11,6 +11,7 @@
 open Names
 open Constr
 open Univ
+open UVars
 open Declarations
 
 (** Unsafe environments. We define here a datatype for environments.
@@ -28,12 +29,6 @@ open Declarations
    - a set of universe constraints
    - a flag telling if Set is, can be, or cannot be set impredicative *)
 
-type lazy_val
-
-val build_lazy_val : lazy_val -> (Vmvalues.values * Id.Set.t) -> unit
-val force_lazy_val : lazy_val -> (Vmvalues.values * Id.Set.t) option
-val dummy_lazy_val : unit -> lazy_val
-
 (** Linking information for the native compiler *)
 type link_info =
   | Linked of string
@@ -60,17 +55,15 @@ end
 
 type named_context_val = private {
   env_named_ctx : Constr.named_context;
-  env_named_map : (Constr.named_declaration * lazy_val) Id.Map.t;
+  env_named_map : Constr.named_declaration Id.Map.t;
   (** Identifier-indexed version of [env_named_ctx] *)
-  env_named_var : Constr.t list;
-  (** List of identifiers in [env_named_ctx], in the same order, including
-      let-ins. This is not used in the kernel, but is critical to preserve
-      sharing of evar instances in the proof engine. *)
+  env_named_idx : Constr.named_declaration Range.t;
+  (** Same as env_named_ctx but with a fast-access list. *)
 }
 
 type rel_context_val = private {
   env_rel_ctx : Constr.rel_context;
-  env_rel_map : (Constr.rel_declaration * lazy_val) Range.t;
+  env_rel_map : Constr.rel_declaration Range.t;
 }
 
 type env = private {
@@ -79,12 +72,23 @@ type env = private {
   env_rel_context   : rel_context_val;
   env_nb_rel        : int;
   env_universes : UGraph.t;
-  env_universes_lbound : UGraph.Bound.t;
+  env_qualities : Sorts.QVar.Set.t;
+  irr_constants : Sorts.relevance Cmap_env.t
+(** [irr_constants] is a cache of the relevances which are not Relevant.
+    In other words, [const_relevance == Option.default Relevant (find_opt con irr_constants)]. *);
+  irr_inds : Sorts.relevance Indmap_env.t
+(** [irr_inds] is a cache of the relevances which are not Relevant. cf [irr_constants]. *);
+  symb_pats : rewrite_rule list Cmap_env.t;
   env_typing_flags  : typing_flags;
+  vm_library : Vmlibrary.t;
   retroknowledge : Retroknowledge.retroknowledge;
-  indirect_pterms : Opaqueproof.opaquetab;
+  rewrite_rules_allowed : bool;
+  (** Allow rewrite rules (breaks e.g. SR) *)
 }
 
+type rewrule_not_allowed = Symb | Rule
+exception RewriteRulesNotAllowed of rewrule_not_allowed
+
 val oracle : env -> Conv_oracle.oracle
 val set_oracle : env -> Conv_oracle.oracle -> env
 
@@ -93,18 +97,12 @@ val eq_named_context_val : named_context_val -> named_context_val -> bool
 val empty_env : env
 
 val universes     : env -> UGraph.t
-val universes_lbound : env -> UGraph.Bound.t
-val set_universes_lbound : env -> UGraph.Bound.t -> env
 val rel_context   : env -> Constr.rel_context
 val named_context : env -> Constr.named_context
 val named_context_val : env -> named_context_val
 
 val set_universes : UGraph.t -> env -> env
 
-val opaque_tables : env -> Opaqueproof.opaquetab
-val set_opaque_tables : env -> Opaqueproof.opaquetab -> env
-
-
 val typing_flags    : env -> typing_flags
 val is_impredicative_set : env -> bool
 val type_in_type : env -> bool
@@ -112,7 +110,6 @@ val deactivated_guard : env -> bool
 val indices_matter : env -> bool
 
 val is_impredicative_sort : env -> Sorts.t -> bool
-val is_impredicative_univ : env -> Univ.Universe.t -> bool
 val is_impredicative_family : env -> Sorts.family -> bool
 
 (** is the local context empty *)
@@ -125,10 +122,13 @@ val push_rel         : Constr.rel_declaration -> env -> env
 val push_rel_context : Constr.rel_context -> env -> env
 val push_rec_types   : rec_declaration -> env -> env
 
+val push_rel_context_val : Constr.rel_declaration -> rel_context_val -> rel_context_val
+val set_rel_context_val : rel_context_val -> env -> env
+val empty_rel_context_val : rel_context_val
+
 (** Looks up in the context of local vars referred by indice ([rel_context])
    raises [Not_found] if the index points out of the context *)
 val lookup_rel    : int -> env -> Constr.rel_declaration
-val lookup_rel_val : int -> env -> lazy_val
 val evaluable_rel : int -> env -> bool
 val env_of_rel     : int -> env -> env
 
@@ -162,7 +162,6 @@ val push_named_context_val  :
    raises [Not_found] if the Id.t is not found *)
 
 val lookup_named     : variable -> env -> Constr.named_declaration
-val lookup_named_val : variable -> env -> lazy_val
 val lookup_named_ctxt : variable -> named_context_val -> Constr.named_declaration
 val evaluable_named  : variable -> env -> bool
 val named_type : variable -> env -> types
@@ -173,7 +172,7 @@ val named_body : variable -> env -> constr option
 val fold_named_context :
   (env -> Constr.named_declaration -> 'a -> 'a) -> env -> init:'a -> 'a
 
-val match_named_context_val : named_context_val -> (named_declaration * lazy_val * named_context_val) option
+val match_named_context_val : named_context_val -> (named_declaration * named_context_val) option
 
 (** Recurrence on [named_context] starting from younger decl *)
 val fold_named_context_reverse :
@@ -207,6 +206,8 @@ val evaluable_constant : Constant.t -> env -> bool
 
 val mem_constant : Constant.t -> env -> bool
 
+val add_rewrite_rules : (Constant.t * rewrite_rule) list -> env -> env
+
 (** New-style polymorphism *)
 val polymorphic_constant  : Constant.t -> env -> bool
 val polymorphic_pconstant : pconstant -> env -> bool
@@ -222,16 +223,17 @@ val type_in_type_constant : Constant.t -> env -> bool
 type const_evaluation_result =
   | NoBody
   | Opaque
-  | IsPrimitive of Univ.Instance.t * CPrimitives.t
+  | IsPrimitive of Instance.t * CPrimitives.t
+  | HasRules of Instance.t * bool * rewrite_rule list
 exception NotEvaluableConst of const_evaluation_result
 
 val constant_type : env -> Constant.t puniverses -> types constrained
 
 val constant_value_and_type : env -> Constant.t puniverses ->
-  constr option * types * Univ.Constraints.t
+  constr option * types * Constraints.t
 (** The universe context associated to the constant, empty if not
     polymorphic *)
-val constant_context : env -> Constant.t -> Univ.AbstractContext.t
+val constant_context : env -> Constant.t -> AbstractContext.t
 
 (* These functions should be called under the invariant that [env]
    already contains the constraints corresponding to the constant
@@ -240,22 +242,26 @@ val constant_value_in : env -> Constant.t puniverses -> constr
 val constant_type_in : env -> Constant.t puniverses -> types
 val constant_opt_value_in : env -> Constant.t puniverses -> constr option
 
+val is_symbol : env -> Constant.t -> bool
 val is_primitive : env -> Constant.t -> bool
 val get_primitive : env -> Constant.t -> CPrimitives.t option
 
 val is_array_type : env -> Constant.t -> bool
 val is_int63_type : env -> Constant.t -> bool
 val is_float64_type : env -> Constant.t -> bool
+val is_string_type : env -> Constant.t -> bool
 val is_primitive_type : env -> Constant.t -> bool
 
 
 (** {6 Primitive projections} *)
 
 (** Checks that the number of parameters is correct. *)
-val lookup_projection : Names.Projection.t -> env -> types
+val lookup_projection : Names.Projection.t -> env -> Sorts.relevance * types
 
-val get_projection : env -> inductive -> proj_arg:int -> Names.Projection.Repr.t option
-val get_projections : env -> inductive -> Names.Projection.Repr.t array option
+(** Anomaly when not a primitive record or invalid proj_arg. *)
+val get_projection : env -> inductive -> proj_arg:int -> Names.Projection.Repr.t * Sorts.relevance
+
+val get_projections : env -> inductive -> (Names.Projection.Repr.t * Sorts.relevance) array option
 
 (** {5 Inductive types } *)
 val lookup_mind_key : MutInd.t -> env -> mind_key
@@ -270,7 +276,7 @@ val mem_mind : MutInd.t -> env -> bool
 
 (** The universe context associated to the inductive, empty if not
     polymorphic *)
-val mind_context : env -> MutInd.t -> Univ.AbstractContext.t
+val mind_context : env -> MutInd.t -> AbstractContext.t
 
 (** New-style polymorphism *)
 val polymorphic_ind  : inductive -> env -> bool
@@ -279,9 +285,32 @@ val type_in_type_ind : inductive -> env -> bool
 
 (** Old-style polymorphism *)
 val template_polymorphic_ind : inductive -> env -> bool
-val template_polymorphic_variables : inductive -> env -> Univ.Level.t list
+val template_polymorphic_variables : inductive -> env -> Level.t list
 val template_polymorphic_pind : pinductive -> env -> bool
 
+(** {6 Changes of representation of Case nodes} *)
+
+(** Given an inductive type and its parameters, builds the context of the return
+    clause, including the inductive being eliminated. The additional binder
+    array is only used to set the names of the context variables, we use the
+    less general type to make it easy to use this function on Case nodes. *)
+val expand_arity : Declarations.mind_specif -> pinductive -> constr array ->
+  Name.t binder_annot array -> rel_context
+
+(** Given an inductive type and its parameters, builds the context of the return
+    clause, including the inductive being eliminated. The additional binder
+    array is only used to set the names of the context variables, we use the
+    less general type to make it easy to use this function on Case nodes. *)
+val expand_branch_contexts : Declarations.mind_specif -> UVars.Instance.t -> constr array ->
+  (Name.t binder_annot array * 'a) array -> rel_context array
+
+(** [instantiate_context u subst nas ctx] applies both [u] and [subst]
+  to [ctx] while replacing names using [nas] (order reversed). In particular,
+  assumes that [ctx] and [nas] have the same length. *)
+val instantiate_context : UVars.Instance.t -> Vars.substl -> Name.t binder_annot array ->
+  rel_context -> rel_context
+
+
 (** {6 Name quotients} *)
 
 module type QNameS =
@@ -290,6 +319,7 @@ sig
   val equal : env -> t -> t -> bool
   val compare : env -> t -> t -> int
   val hash : env -> t -> int
+  val canonize : env -> t -> t
 end
 
 module QConstant : QNameS with type t = Constant.t
@@ -320,23 +350,26 @@ val lookup_modtype : ModPath.t -> env -> module_type_body
 
 (** {5 Universe constraints } *)
 
-val add_constraints : Univ.Constraints.t -> env -> env
+val add_constraints : Constraints.t -> env -> env
 (** Add universe constraints to the environment.
     @raise UniverseInconsistency. *)
 
-val check_constraints : Univ.Constraints.t -> env -> bool
+val check_constraints : Constraints.t -> env -> bool
 (** Check constraints are satifiable in the environment. *)
 
-val push_context : ?strict:bool -> Univ.UContext.t -> env -> env
+val push_context : ?strict:bool -> UContext.t -> env -> env
 (** [push_context ?(strict=false) ctx env] pushes the universe context to the environment.
     @raise UGraph.AlreadyDeclared if one of the universes is already declared. *)
 
-val push_context_set : ?strict:bool -> Univ.ContextSet.t -> env -> env
+val push_context_set : ?strict:bool -> ContextSet.t -> env -> env
 (** [push_context_set ?(strict=false) ctx env] pushes the universe
     context set to the environment. It does not fail even if one of the
     universes is already declared. *)
 
-val push_subgraph : Univ.ContextSet.t -> env -> env
+val push_floating_context_set : ContextSet.t -> env -> env
+(** Same as above but keep the universes floating for template. Do not use. *)
+
+val push_subgraph : ContextSet.t -> env -> env
 (** [push_subgraph univs env] adds the universes and constraints in
    [univs] to [env] as [push_context_set ~strict:false univs env], and
    also checks that they do not imply new transitive constraints
@@ -344,10 +377,11 @@ val push_subgraph : Univ.ContextSet.t -> env -> env
 
 val set_typing_flags : typing_flags -> env -> env
 val set_impredicative_set : bool -> env -> env
-val set_cumulative_sprop : bool -> env -> env
 val set_type_in_type : bool -> env -> env
 val set_allow_sprop : bool -> env -> env
 val sprop_allowed : env -> bool
+val allow_rewrite_rules : env -> env
+val rewrite_rules_allowed : env -> bool
 
 val same_flags : typing_flags -> typing_flags -> bool
 
@@ -390,11 +424,11 @@ val make_judge : 'constr -> 'types -> ('constr, 'types) punsafe_judgment
 val j_val  : ('constr, 'types) punsafe_judgment -> 'constr
 val j_type : ('constr, 'types) punsafe_judgment -> 'types
 
-type 'types punsafe_type_judgment = {
+type ('types, 'sorts) punsafe_type_judgment = {
   utj_val : 'types;
-  utj_type : Sorts.t }
+  utj_type : 'sorts }
 
-type unsafe_type_judgment = types punsafe_type_judgment
+type unsafe_type_judgment = (types, Sorts.t) punsafe_type_judgment
 
 exception Hyp_not_found
 
@@ -405,13 +439,20 @@ val apply_to_hyp : named_context_val -> variable ->
   (Constr.named_context -> Constr.named_declaration -> Constr.named_context -> Constr.named_declaration) ->
     named_context_val
 
-val remove_hyps : Id.Set.t -> (Constr.named_declaration -> Constr.named_declaration) -> (lazy_val -> lazy_val) -> named_context_val -> named_context_val
+val remove_hyps : Id.Set.t -> (Constr.named_declaration -> Constr.named_declaration) -> named_context_val -> named_context_val
 
 val is_polymorphic : env -> Names.GlobRef.t -> bool
 val is_template_polymorphic : env -> GlobRef.t -> bool
-val get_template_polymorphic_variables : env -> GlobRef.t -> Univ.Level.t list
+val get_template_polymorphic_variables : env -> GlobRef.t -> Level.t list
 val is_type_in_type : env -> GlobRef.t -> bool
 
+(** {5 VM and native} *)
+
+val vm_library : env -> Vmlibrary.t
+val set_vm_library : Vmlibrary.t -> env -> env
+val link_vm_library : Vmlibrary.on_disk -> env -> env
+val lookup_vm_code : Vmlibrary.index -> env -> Vmemitcodes.to_patch
+
 (** Native compiler *)
 val no_link_info : link_info
 
