@@ -25,7 +25,7 @@ open Constr
    Equiv gives the canonical name in the given context. *)
 
 type delta_hint =
-  | Inline of int * constr Univ.univ_abstracted option
+  | Inline of int * constr UVars.univ_abstracted option
   | Equiv of KerName.t
 
 (* NB: earlier constructor Prefix_equiv of ModPath.t
@@ -165,7 +165,7 @@ let find_prefix resolve mp =
 
 (** Applying a resolver to a kernel name *)
 
-exception Change_equiv_to_inline of (int * constr Univ.univ_abstracted)
+exception Change_equiv_to_inline of (int * constr UVars.univ_abstracted)
 
 let solve_delta_kn resolve kn =
   try
@@ -306,16 +306,10 @@ let subst_con0 subst cst =
   let mpu,mpc,resolve,user = subst_dual_mp subst mpu mpc in
   let knu = KerName.make mpu l in
   let knc = if mpu == mpc then knu else KerName.make mpc l in
-  match search_delta_inline resolve knu knc with
-    | Some t ->
-      (* In case of inlining, discard the canonical part (cf #2608) *)
-      Constant.make1 knu, Some t
-    | None ->
-      let knc' =
-        progress (kn_of_delta resolve) (if user then knu else knc) ~orelse:knc
-      in
-      let cst' = Constant.make knu knc' in
-      cst', None
+  let knc' =
+    progress (kn_of_delta resolve) (if user then knu else knc) ~orelse:knc in
+  let cst' = Constant.make knu knc' in
+  cst', search_delta_inline resolve knu knc
 
 let subst_con subst cst =
   try subst_con0 subst cst
@@ -350,18 +344,18 @@ let rec map_kn f f' c =
   let func = map_kn f f' in
     match kind c with
       | Const kn -> (try f' kn with No_subst -> c)
-      | Proj (p,t) ->
+      | Proj (p,r,t) ->
           let p' = Projection.map f p in
           let t' = func t in
             if p' == p && t' == t then c
-            else mkProj (p', t')
+            else mkProj (p', r, t')
       | Ind ((kn,i),u) ->
           let kn' = f kn in
           if kn'==kn then c else mkIndU ((kn',i),u)
       | Construct (((kn,i),j),u) ->
           let kn' = f kn in
           if kn'==kn then c else mkConstructU (((kn',i),j),u)
-      | Case (ci,u,pms,p,iv,ct,l) ->
+      | Case (ci,u,pms,(p,r),iv,ct,l) ->
           let ci_ind =
             let (kn,i) = ci.ci_ind in
             let kn' = f kn in
@@ -380,7 +374,7 @@ let rec map_kn f f' c =
                 && l'==l && ct'==ct)then c
             else
               mkCase ({ci with ci_ind = ci_ind}, u,
-                      pms',p',iv',ct', l')
+                      pms',(p',r),iv',ct', l')
       | Cast (ct,k,t) ->
           let ct' = func ct in
           let t'= func t in
@@ -433,6 +427,41 @@ let subst_mps subst c =
   if is_empty_subst subst then c
   else map_kn (subst_mind subst) (subst_pcon_term subst) c
 
+let subst_mps_aux subst = function
+| Inl (con, u) ->
+  begin match subst_con0 subst con with
+  | con', None -> Inl (con', u)
+  | _, Some t -> Inr (Vars.univ_instantiate_constr u t)
+  | exception No_subst -> Inl (con, u)
+  end
+| Inr t -> Inr (subst_mps subst t)
+
+let subst_mps_list substs c =
+  if List.is_empty substs || List.for_all is_empty_subst substs then c
+  else
+    let cache_const = ref Cmap_env.empty in
+    let cache_ind = ref Mindmap_env.empty in
+    let subst_const (con, u as pcon) = match Cmap_env.find_opt con !cache_const with
+    | Some ans ->
+      if ans == con then raise No_subst else mkConstU (ans, u)
+    | None ->
+      let ans = List.fold_right subst_mps_aux substs (Inl pcon) in
+      (* Do not cache arbitrary inline terms *)
+      match ans with
+      | Inl (con', _ as ans) ->
+        let () = cache_const := Cmap_env.add con con' !cache_const in
+        if con' == con then raise No_subst else mkConstU ans
+      | Inr ans -> ans
+    in
+    let subst_mind ind = match Mindmap_env.find_opt ind !cache_ind with
+    | Some ans -> ans
+    | None ->
+      let ans = List.fold_right subst_mind substs ind in
+      let () = cache_ind := Mindmap_env.add ind ans !cache_ind in
+      ans
+    in
+    map_kn subst_mind subst_const c
+
 let rec replace_mp_in_mp mpfrom mpto mp =
   match mp with
     | _ when ModPath.equal mp mpfrom -> mpto
@@ -496,7 +525,7 @@ let gen_subst_delta_resolver dom subst resolver =
       | Equiv kequ ->
           (try Equiv (subst_kn_delta subst kequ)
            with Change_equiv_to_inline (lev,c) -> Inline (lev,Some c))
-      | Inline (lev,Some t) -> Inline (lev,Some (Univ.map_univ_abstracted (subst_mps subst) t))
+      | Inline (lev,Some t) -> Inline (lev,Some (UVars.map_univ_abstracted (subst_mps subst) t))
       | Inline (_,None) -> hint
     in
     Deltamap.add_kn kkey' hint' rslv
