@@ -11,12 +11,44 @@
 open Util
 open CErrors
 open Names
+open Values
 open Constr
 
 (** This module defines the representation of values internally used by
 the native compiler *)
 
-type t = t -> t
+type ('a,'b) eq = ('a,'b) Util.eq = Refl : ('a,'a) eq
+
+type t
+let t_eq : (t, t -> t) eq = Obj.magic ()
+let t_eq2 : (t, t -> t -> t) eq = Obj.magic ()
+let t_eq3 : (t, t -> t -> t -> t) eq = Obj.magic ()
+let t_eq4 : (t, t -> t -> t -> t -> t) eq = Obj.magic ()
+
+let cast_gen (type a) (type b) (e:(a,b) eq) (x:a) : b =
+  let Refl = e in x
+
+let apply (f:t) : t -> t =
+  cast_gen t_eq f
+
+(* Writing this as [apply2 f x y = apply (apply f x) y] seems to
+   confuse ocaml and it produces less efficient code.
+
+   When written with a direct [cast_gen t_eq2] ocaml seems smart
+   enough to fully reduce it where used. *)
+let apply2 (f:t) : t -> t -> t =
+  cast_gen t_eq2 f
+
+let apply3 (f:t) : t -> t -> t -> t =
+  cast_gen t_eq3 f
+
+let apply4 (f:t) : t -> t -> t -> t -> t =
+  cast_gen t_eq4 f
+
+let of_fun (f:t->t) : t =
+  cast_gen (sym t_eq) f
+
+let eta_expand f = of_fun (fun x -> apply f x)
 
 type accumulator (* = t (* a block [0:code;atom;arguments] *) *)
 
@@ -28,7 +60,6 @@ type reloc_table = (tag * arity) array
 
 type annot_sw = {
     asw_ind : inductive;
-    asw_ci : case_info;
     asw_reloc : reloc_table;
     asw_finite : bool;
     asw_prefix : string
@@ -50,19 +81,18 @@ type rec_pos = int array
 
 let eq_rec_pos = Array.equal Int.equal
 
+type vcofix = CofixLazy of t | CofixValue of t
+
 type atom =
   | Arel of int
   | Aconstant of pconstant
   | Aind of pinductive
   | Asort of Sorts.t
   | Avar of Id.t
-  | Acase of annot_sw * accumulator * t * (t -> t)
+  | Acase of annot_sw * accumulator * t * t
   | Afix of t array * t array * rec_pos * int
             (* types, bodies, rec_pos, pos *)
-  | Acofix of t array * t array * int * t
-  | Acofixe of t array * t array * int * t
-  | Aprod of Name.t * t * (t -> t)
-  | Ameta of metavariable * t
+  | Acofix of t array * t array * int * vcofix
   | Aevar of Evar.t * t array
   | Aproj of (inductive * int) * accumulator
 
@@ -73,11 +103,12 @@ type symbol =
   | SymbConst of Constant.t
   | SymbMatch of annot_sw
   | SymbInd of inductive
-  | SymbMeta of metavariable
   | SymbEvar of Evar.t
-  | SymbLevel of Univ.Level.t
+  | SymbInstance of UVars.Instance.t
   | SymbProj of (inductive * int)
 
+type block
+
 type symbols = symbol array
 
 let empty_symbols = [| |]
@@ -88,22 +119,24 @@ let accumulate_tag = 0
 (** Unique pointer used to drive the accumulator function *)
 let ret_accu = Obj.repr (ref ())
 
-type accu_val = { mutable acc_atm : atom; acc_arg : Obj.t list }
+type accu_val = { mutable acc_atm : atom; acc_arg : t list }
+
+external set_tag : Obj.t -> int -> unit = "coq_obj_set_tag"
 
 let mk_accu (a : atom) : t =
   let rec accumulate data x =
-    if x == ret_accu then Obj.repr data
+    if Obj.repr x == ret_accu then Obj.repr data
     else
       let data = { data with acc_arg = x :: data.acc_arg } in
       let ans = Obj.repr (accumulate data) in
-      let () = Obj.set_tag ans accumulate_tag [@ocaml.warning "-3"] in
+      let () = set_tag ans accumulate_tag in
       ans
   in
   let acc = { acc_atm = a; acc_arg = [] } in
   let ans = Obj.repr (accumulate acc) in
   (** FIXME: use another representation for accumulators, this causes naked
       pointers. *)
-  let () = Obj.set_tag ans accumulate_tag [@ocaml.warning "-3"] in
+  let () = set_tag ans accumulate_tag in
   (Obj.obj ans : t)
 
 let get_accu (k : accumulator) =
@@ -134,22 +167,15 @@ let mk_rels_accu lvl len =
   Array.init len (fun i -> mk_rel_accu (lvl + i))
 
 let napply (f:t) (args: t array) =
-  Array.fold_left (fun f a -> f a) f args
+  Array.fold_left (fun f a -> apply f a) f args
 
-let mk_constant_accu kn u =
-  mk_accu (Aconstant (kn,Univ.Instance.of_array u))
+let mk_constant_accu kn u = mk_accu (Aconstant (kn,u))
 
-let mk_ind_accu ind u =
-  mk_accu (Aind (ind,Univ.Instance.of_array u))
+let mk_ind_accu ind u = mk_accu (Aind (ind,u))
 
 let mk_sort_accu s u =
-  let open Sorts in
-  match s with
-  | SProp | Prop | Set -> mk_accu (Asort s)
-  | Type s ->
-     let u = Univ.Instance.of_array u in
-     let s = Sorts.sort_of_univ (Univ.subst_instance_universe u s) in
-     mk_accu (Asort s)
+  let s = UVars.subst_instance_sort u s in
+  mk_accu (Asort s)
 
 let mk_var_accu id =
   mk_accu (Avar id)
@@ -157,11 +183,18 @@ let mk_var_accu id =
 let mk_sw_accu annot c p ac =
   mk_accu (Acase(annot,c,p,ac))
 
-let mk_prod_accu s dom codom =
-  mk_accu (Aprod (s,dom,codom))
+let prod_tag =
+  (* We rely on the tag of Vprod! *)
+  let () = assert (Obj.tag (Obj.repr (Vprod (Anonymous, Obj.magic 0, Obj.magic 0))) == 2) in
+  2
 
-let mk_meta_accu mv ty =
-  mk_accu (Ameta (mv,ty))
+let mk_prod s dom codom =
+  (* [Prod (s, dom, codom)] is coded as [tag:0|[tag:2|s; dom; codom]]
+     This looks like a PArray but has a tag distinct from all PArray values on
+     the inner block. This cannot be an accumulator because all accumulators
+     have length >= 2. *)
+  let block = Obj.repr (Vprod (s, dom, codom)) in
+  (Obj.magic (ref block) : t)
 
 let mk_evar_accu ev args =
   mk_accu (Aevar (ev, args))
@@ -179,33 +212,32 @@ let accu_nargs (k:accumulator) =
   List.length (get_accu k).acc_arg
 
 let args_of_accu (k:accumulator) =
-  let acc = (get_accu k).acc_arg in
-  (Obj.magic (Array.of_list acc) : t array)
+  (get_accu k).acc_arg
 
 let mk_fix_accu rec_pos pos types bodies =
   mk_accu (Afix(types,bodies,rec_pos, pos))
 
 let mk_cofix_accu pos types norm =
-  mk_accu (Acofix(types,norm,pos,(Obj.magic 0 : t)))
+  mk_accu (Acofix (types, norm, pos, CofixLazy (Obj.magic 0 : t)))
 
 let upd_cofix (cofix :t) (cofix_fun : t) =
   let atom = atom_of_accu (Obj.magic cofix) in
   match atom with
   | Acofix (typ,norm,pos,_) ->
-      set_atom_of_accu (Obj.magic cofix) (Acofix(typ,norm,pos,cofix_fun))
+    set_atom_of_accu (Obj.magic cofix) (Acofix (typ, norm, pos, CofixLazy cofix_fun))
   | _ -> assert false
 
 let force_cofix (cofix : t) =
   let accu = (Obj.magic cofix : accumulator) in
   let atom = atom_of_accu accu in
   match atom with
-  | Acofix(typ,norm,pos,f) ->
+  | Acofix (typ, norm, pos, CofixLazy f) ->
     let args = args_of_accu accu in
-    let f = Array.fold_right (fun arg f -> f arg) args f in
-    let v = f (Obj.magic ()) in
-    set_atom_of_accu accu (Acofixe(typ,norm,pos,v));
+    let f = List.fold_right (fun arg f -> apply f arg) args f in
+    let v = apply f (Obj.magic ()) in
+    let () = set_atom_of_accu accu (Acofix (typ, norm, pos, CofixValue v)) in
       v
-  | Acofixe(_,_,_,v) -> v
+  | Acofix (_, _, _, CofixValue v) -> v
   | _ -> cofix
 
 let mk_const tag = Obj.magic tag
@@ -221,7 +253,7 @@ let mk_block tag args =
 (* Two instances of dummy_value should not be pointer equal, otherwise
  comparing them as terms would succeed *)
 let dummy_value : unit -> t =
-  fun () _ -> anomaly ~label:"native" (Pp.str "Evaluation failed.")
+  fun () -> of_fun (fun _ -> anomaly ~label:"native" (Pp.str "Evaluation failed."))
 
 let cast_accu v = (Obj.magic v:accumulator)
 [@@ocaml.inline always]
@@ -239,7 +271,8 @@ let mk_uint (x : Uint63.t) = (Obj.magic x : t)
 let mk_float (x : Float64.t) = (Obj.magic x : t)
 [@@ocaml.inline always]
 
-type block
+let mk_string (x : Pstring.t) = (Obj.magic x : t)
+[@@ocaml.inline always]
 
 let block_size (b:block) =
   Obj.size (Obj.magic b)
@@ -249,31 +282,28 @@ let block_field (b:block) i = (Obj.magic (Obj.field (Obj.magic b) i) : t)
 let block_tag (b:block) =
   Obj.tag (Obj.magic b)
 
-type kind_of_value =
-  | Vaccu of accumulator
-  | Vfun of (t -> t)
-  | Vconst of int
-  | Vint64 of int64
-  | Vfloat64 of float
-  | Varray of t Parray.t
-  | Vblock of block
+type kind = (t, accumulator, t -> t, Name.t * t * t, Empty.t, Empty.t, block) Values.kind
 
 let kind_of_value (v:t) =
   let o = Obj.repr v in
   if Obj.is_int o then Vconst (Obj.magic v)
-  else if Obj.tag o == Obj.double_tag then Vfloat64 (Obj.magic v)
   else
     let tag = Obj.tag o in
     if Int.equal tag accumulate_tag then
-      if Int.equal (Obj.size o) 1 then Varray (Obj.magic v)
+      if Int.equal (Obj.size o) 1 then
+        let w = Obj.field o 0 in
+        let tag = Obj.tag w in
+        if Int.equal tag prod_tag then Obj.magic w
+        else Varray (Obj.magic v)
       else Vaccu (Obj.magic v)
     else if Int.equal tag Obj.custom_tag then Vint64 (Obj.magic v)
     else if Int.equal tag Obj.double_tag then Vfloat64 (Obj.magic v)
+    else if Int.equal tag Obj.string_tag then Vstring (Obj.magic v)
     else if (tag < Obj.lazy_tag) then Vblock (Obj.magic v)
       else
         (* assert (tag = Obj.closure_tag || tag = Obj.infix_tag);
            or ??? what is 1002*)
-        Vfun v
+        Vfun (apply v)
 
 (** Support for machine integers *)
 
@@ -294,7 +324,7 @@ let no_check_head0 x =
 
 let head0 accu x =
  if is_int x then  no_check_head0 x
- else accu x
+ else apply accu x
 
 let no_check_tail0 x =
   mk_uint (Uint63.tail0 (to_uint x))
@@ -302,7 +332,7 @@ let no_check_tail0 x =
 
 let tail0 accu x =
  if is_int x then no_check_tail0 x
- else accu x
+ else apply accu x
 
 let no_check_add  x y =
   mk_uint (Uint63.add (to_uint x) (to_uint y))
@@ -310,7 +340,7 @@ let no_check_add  x y =
 
 let add accu x y =
   if is_int x && is_int y then no_check_add x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_sub x y =
   mk_uint (Uint63.sub (to_uint x) (to_uint y))
@@ -318,7 +348,7 @@ let no_check_sub x y =
 
 let sub accu x y =
   if is_int x && is_int y then no_check_sub x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_mul x y =
   mk_uint (Uint63.mul (to_uint x) (to_uint y))
@@ -326,7 +356,7 @@ let no_check_mul x y =
 
 let mul accu x y =
   if is_int x && is_int y then no_check_mul x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_div x y =
   mk_uint (Uint63.div (to_uint x) (to_uint y))
@@ -334,7 +364,7 @@ let no_check_div x y =
 
 let div accu x y =
   if is_int x && is_int y then no_check_div x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_rem x y =
   mk_uint (Uint63.rem (to_uint x) (to_uint y))
@@ -342,7 +372,7 @@ let no_check_rem x y =
 
 let rem accu x y =
   if is_int x && is_int y then no_check_rem x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_divs x y =
   mk_uint (Uint63.divs (to_uint x) (to_uint y))
@@ -350,7 +380,7 @@ let no_check_divs x y =
 
 let divs accu x y =
   if is_int x && is_int y then no_check_divs x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_rems x y =
   mk_uint (Uint63.rems (to_uint x) (to_uint y))
@@ -358,7 +388,7 @@ let no_check_rems x y =
 
 let rems accu x y =
   if is_int x && is_int y then no_check_rems x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_sr x y =
   mk_uint (Uint63.l_sr (to_uint x) (to_uint y))
@@ -366,7 +396,7 @@ let no_check_l_sr x y =
 
 let l_sr accu x y =
   if is_int x && is_int y then no_check_l_sr x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_sl x y =
   mk_uint (Uint63.l_sl (to_uint x) (to_uint y))
@@ -374,7 +404,7 @@ let no_check_l_sl x y =
 
 let l_sl accu x y =
   if is_int x && is_int y then no_check_l_sl x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_a_sr x y =
   mk_uint (Uint63.a_sr (to_uint x) (to_uint y))
@@ -382,7 +412,7 @@ let no_check_a_sr x y =
 
 let a_sr accu x y =
   if is_int x && is_int y then no_check_a_sr x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_and x y =
   mk_uint (Uint63.l_and (to_uint x) (to_uint y))
@@ -390,7 +420,7 @@ let no_check_l_and x y =
 
 let l_and accu x y =
   if is_int x && is_int y then no_check_l_and x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_xor x y =
   mk_uint (Uint63.l_xor (to_uint x) (to_uint y))
@@ -398,7 +428,7 @@ let no_check_l_xor x y =
 
 let l_xor accu x y =
   if is_int x && is_int y then no_check_l_xor x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_or x y =
   mk_uint (Uint63.l_or (to_uint x) (to_uint y))
@@ -406,7 +436,7 @@ let no_check_l_or x y =
 
 let l_or accu x y =
   if is_int x && is_int y then no_check_l_or x y
-  else accu x y
+  else apply2 accu x y
 
 [@@@ocaml.warning "-37"]
 type coq_carry =
@@ -429,7 +459,7 @@ let no_check_addc x y =
 
 let addc accu x y =
   if is_int x && is_int y then no_check_addc x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_subc x y =
   let s = Uint63.sub (to_uint x) (to_uint y) in
@@ -438,7 +468,7 @@ let no_check_subc x y =
 
 let subc accu x y =
   if is_int x && is_int y then no_check_subc x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_addCarryC x y =
   let s =
@@ -449,7 +479,7 @@ let no_check_addCarryC x y =
 
 let addCarryC accu x y =
   if is_int x && is_int y then no_check_addCarryC x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_subCarryC x y =
   let s =
@@ -460,7 +490,7 @@ let no_check_subCarryC x y =
 
 let subCarryC accu x y =
   if is_int x && is_int y then no_check_subCarryC x y
-  else accu x y
+  else apply2 accu x y
 
 let of_pair (x, y) =
   (Obj.magic (PPair(mk_uint x, mk_uint y)):t)
@@ -472,7 +502,7 @@ let no_check_mulc x y =
 
 let mulc accu x y =
   if is_int x && is_int y then no_check_mulc x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_diveucl x y =
   let i1, i2 = to_uint x, to_uint y in
@@ -481,7 +511,7 @@ let no_check_diveucl x y =
 
 let diveucl accu x y =
   if is_int x && is_int y then no_check_diveucl x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_div21 x y z =
   let i1, i2, i3 = to_uint x, to_uint y, to_uint z in
@@ -490,7 +520,7 @@ let no_check_div21 x y z =
 
 let div21 accu x y z =
   if is_int x && is_int y && is_int z then no_check_div21 x y z
-  else accu x y z
+  else apply3 accu x y z
 
 let no_check_addMulDiv x y z =
   let p, i, j = to_uint x, to_uint y, to_uint z in
@@ -499,7 +529,7 @@ let no_check_addMulDiv x y z =
 
 let addMulDiv accu x y z =
   if is_int x && is_int y && is_int z then no_check_addMulDiv x y z
-  else accu x y z
+  else apply3 accu x y z
 
 [@@@ocaml.warning "-34"]
 type coq_bool =
@@ -519,7 +549,7 @@ let no_check_eq x y =
 
 let eq accu x y =
   if is_int x && is_int y then no_check_eq x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_lt x y =
   mk_bool (Uint63.lt (to_uint x) (to_uint y))
@@ -527,7 +557,7 @@ let no_check_lt x y =
 
 let lt accu x y =
   if is_int x && is_int y then no_check_lt x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_le x y =
   mk_bool (Uint63.le (to_uint x) (to_uint y))
@@ -535,7 +565,7 @@ let no_check_le x y =
 
 let le accu x y =
   if is_int x && is_int y then no_check_le x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_lts x y =
   mk_bool (Uint63.lts (to_uint x) (to_uint y))
@@ -543,7 +573,7 @@ let no_check_lts x y =
 
 let lts accu x y =
   if is_int x && is_int y then no_check_lts x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_les x y =
   mk_bool (Uint63.les (to_uint x) (to_uint y))
@@ -551,7 +581,7 @@ let no_check_les x y =
 
 let les accu x y =
   if is_int x && is_int y then no_check_les x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_compare x y =
   match Uint63.compare (to_uint x) (to_uint y) with
@@ -561,7 +591,7 @@ let no_check_compare x y =
 
 let compare accu x y =
   if is_int x && is_int y then no_check_compare x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_compares x y =
   match Uint63.compares (to_uint x) (to_uint y) with
@@ -571,7 +601,7 @@ let no_check_compares x y =
 
 let compares accu x y =
   if is_int x && is_int y then no_check_compares x y
-  else accu x y
+  else apply2 accu x y
 
 let print x =
   Printf.fprintf stderr "%s" (Uint63.to_string (to_uint x));
@@ -592,7 +622,7 @@ let no_check_fopp x =
 
 let fopp accu x =
   if is_float x then no_check_fopp x
-  else accu x
+  else apply accu x
 
 let no_check_fabs x =
   mk_float (Float64.abs (to_float x))
@@ -600,28 +630,31 @@ let no_check_fabs x =
 
 let fabs accu x =
   if is_float x then no_check_fabs x
-  else accu x
+  else apply accu x
 
 let no_check_feq x y =
   mk_bool (Float64.eq (to_float x) (to_float y))
+[@@ocaml.inline always]
 
 let feq accu x y =
   if is_float x && is_float y then no_check_feq x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_flt x y =
   mk_bool (Float64.lt (to_float x) (to_float y))
+[@@ocaml.inline always]
 
 let flt accu x y =
   if is_float x && is_float y then no_check_flt x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fle x y =
   mk_bool (Float64.le (to_float x) (to_float y))
+[@@ocaml.inline always]
 
 let fle accu x y =
   if is_float x && is_float y then no_check_fle x y
-  else accu x y
+  else apply2 accu x y
 
 type coq_fcmp =
   | CFcmpAccu of t
@@ -637,14 +670,15 @@ let no_check_fcompare x y =
 
 let fcompare accu x y =
   if is_float x && is_float y then no_check_fcompare x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fequal x y =
   mk_bool (Float64.equal (to_float x) (to_float y))
+[@@ocaml.inline always]
 
 let fequal accu x y =
   if is_float x && is_float y then no_check_fequal x y
-  else accu x y
+  else apply2 accu x y
 
 type coq_fclass =
   | CFclassAccu of t
@@ -665,7 +699,7 @@ let no_check_fclassify x =
 
 let fclassify accu x =
   if is_float x then no_check_fclassify x
-  else accu x
+  else apply accu x
 
 let no_check_fadd x y =
   mk_float (Float64.add (to_float x) (to_float y))
@@ -673,7 +707,7 @@ let no_check_fadd x y =
 
 let fadd accu x y =
   if is_float x && is_float y then no_check_fadd x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fsub x y =
   mk_float (Float64.sub (to_float x) (to_float y))
@@ -681,7 +715,7 @@ let no_check_fsub x y =
 
 let fsub accu x y =
   if is_float x && is_float y then no_check_fsub x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fmul x y =
   mk_float (Float64.mul (to_float x) (to_float y))
@@ -689,7 +723,7 @@ let no_check_fmul x y =
 
 let fmul accu x y =
   if is_float x && is_float y then no_check_fmul x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fdiv x y =
   mk_float (Float64.div (to_float x) (to_float y))
@@ -697,7 +731,7 @@ let no_check_fdiv x y =
 
 let fdiv accu x y =
   if is_float x && is_float y then no_check_fdiv x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fsqrt x =
   mk_float (Float64.sqrt (to_float x))
@@ -705,7 +739,7 @@ let no_check_fsqrt x =
 
 let fsqrt accu x =
   if is_float x then no_check_fsqrt x
-  else accu x
+  else apply accu x
 
 let no_check_float_of_int x =
   mk_float (Float64.of_uint63 (to_uint x))
@@ -713,7 +747,7 @@ let no_check_float_of_int x =
 
 let float_of_int accu x =
   if is_int x then no_check_float_of_int x
-  else accu x
+  else apply accu x
 
 let no_check_normfr_mantissa x =
   mk_uint (Float64.normfr_mantissa (to_float x))
@@ -721,7 +755,7 @@ let no_check_normfr_mantissa x =
 
 let normfr_mantissa accu x =
   if is_float x then no_check_normfr_mantissa x
-  else accu x
+  else apply accu x
 
 let no_check_frshiftexp x =
   let f, e = Float64.frshiftexp (to_float x) in
@@ -730,7 +764,7 @@ let no_check_frshiftexp x =
 
 let frshiftexp accu x =
   if is_float x then no_check_frshiftexp x
-  else accu x
+  else apply accu x
 
 let no_check_ldshiftexp x e =
   mk_float (Float64.ldshiftexp (to_float x) (to_uint e))
@@ -738,7 +772,7 @@ let no_check_ldshiftexp x e =
 
 let ldshiftexp accu x e =
   if is_float x && is_int e then no_check_ldshiftexp x e
-  else accu x e
+  else apply2 accu x e
 
 let no_check_next_up x =
   mk_float (Float64.next_up (to_float x))
@@ -746,7 +780,7 @@ let no_check_next_up x =
 
 let next_up accu x =
   if is_float x then no_check_next_up x
-  else accu x
+  else apply accu x
 
 let no_check_next_down x =
   mk_float (Float64.next_down (to_float x))
@@ -754,9 +788,85 @@ let no_check_next_down x =
 
 let next_down accu x =
   if is_float x then no_check_next_down x
-  else accu x
+  else apply accu x
+
+(** Support for primitive strings *)
+
+let is_string (x:t) =
+  let o = Obj.repr x in
+  Int.equal (Obj.tag o) Obj.string_tag
+
+let to_string (x:t) = (Obj.magic x : Pstring.t)
+[@@ocaml.inline always]
+
+let no_check_string_make n c =
+  mk_string (Pstring.make (to_uint n) (to_uint c))
+[@@ocaml.inline always]
+
+let no_check_string_length s =
+  mk_uint (Pstring.length (to_string s))
+[@@ocaml.inline always]
+
+let no_check_string_get s i =
+  mk_uint (Pstring.get (to_string s) (to_uint i))
+[@@ocaml.inline always]
+
+let no_check_string_sub s off len =
+  mk_string (Pstring.sub (to_string s) (to_uint off) (to_uint len))
+[@@ocaml.inline always]
+
+let no_check_string_cat s1 s2 =
+  mk_string (Pstring.cat (to_string s1) (to_string s2))
+[@@ocaml.inline always]
+
+let no_check_string_compare s1 s2 =
+  match Pstring.compare (to_string s1) (to_string s2) with
+  | x when x < 0 -> (Obj.magic CmpLt:t)
+  | 0 -> (Obj.magic CmpEq:t)
+  | _ -> (Obj.magic CmpGt:t)
+
+let string_make accu n c =
+  if is_int n && is_int c then
+    no_check_string_make n c
+  else
+    apply2 accu n c
+
+let string_length accu s =
+  if is_string s then
+    no_check_string_length s
+  else
+    apply accu s
+
+let string_get accu s i =
+  if is_string s && is_int i then
+    no_check_string_get s i
+  else
+    apply2 accu s i
+
+let string_sub accu s off len =
+  if is_string s && is_int off && is_int len then
+    no_check_string_sub s off len
+  else
+    apply3 accu s off len
+
+let string_cat accu s1 s2 =
+  if is_string s1 && is_string s2 then
+    no_check_string_cat s1 s2
+  else
+    apply2 accu s1 s2
+
+let string_compare accu s1 s2 =
+  if is_string s1 && is_string s2 then
+    no_check_string_compare s1 s2
+  else
+    apply2 accu s1 s2
+
+(** Support for primitive arrays *)
 
 let is_parray t =
+  (* This is only used over values known to inhabit an array type, so we just
+     have to discriminate between actual arrays and accumulators. The latter
+     are always closures with the tag set to 0, so they have size >= 2. *)
   let t = Obj.magic t in
   Obj.is_block t && Obj.size t = 1
 
@@ -765,11 +875,12 @@ let of_parray t = Obj.magic t
 
 let no_check_arraymake n def =
   of_parray (Parray.make (to_uint n) def)
+[@@ocaml.inline always]
 
 let arraymake accu vA n def =
   if is_int n then
     no_check_arraymake n def
-  else accu vA n def
+  else apply3 accu vA n def
 
 let no_check_arrayget t n =
    Parray.get (to_parray t) (to_uint n)
@@ -778,7 +889,7 @@ let no_check_arrayget t n =
 let arrayget accu vA t n =
   if is_parray t && is_int n then
     no_check_arrayget t n
-  else accu vA t n
+  else apply3 accu vA t n
 
 let no_check_arraydefault t =
   Parray.default (to_parray t)
@@ -787,7 +898,7 @@ let no_check_arraydefault t =
 let arraydefault accu vA t =
   if is_parray t then
     no_check_arraydefault t
-  else accu vA t
+  else apply2 accu vA t
 
 let no_check_arrayset t n v =
   of_parray (Parray.set (to_parray t) (to_uint n) v)
@@ -796,7 +907,7 @@ let no_check_arrayset t n v =
 let arrayset accu vA t n v =
   if is_parray t && is_int n then
     no_check_arrayset t n v
-  else accu vA t n v
+  else apply4 accu vA t n v
 
 let no_check_arraycopy t =
   of_parray (Parray.copy (to_parray t))
@@ -805,7 +916,7 @@ let no_check_arraycopy t =
 let arraycopy accu vA t =
   if is_parray t then
     no_check_arraycopy t
-  else accu vA t
+  else apply2 accu vA t
 
 let no_check_arraylength t =
   mk_uint (Parray.length (to_parray t))
@@ -814,17 +925,11 @@ let no_check_arraylength t =
 let arraylength accu vA t =
   if is_parray t then
     no_check_arraylength t
-  else accu vA t
+  else apply2 accu vA t
 
 let parray_of_array t def =
   (Obj.magic (Parray.unsafe_of_obj (Obj.repr t) def) : t)
 
-let arrayinit n (f:t->t) def =
-  of_parray (Parray.init (to_uint n) (Obj.magic f) def)
-
-let arraymap f t =
-  of_parray (Parray.map f (to_parray t))
-
 let hobcnv = Array.init 256 (fun i -> Printf.sprintf "%02x" i)
 let bohcnv = Array.init 256 (fun i -> i -
                                       (if 0x30 <= i then 0x30 else 0) -
