@@ -12,7 +12,6 @@ open Names
 
 type vodigest =
   | Dvo_or_vi of Digest.t        (* The digest of the seg_lib part *)
-  | Dvivo of Digest.t * Digest.t (* The digest of the seg_lib + seg_univ part *)
 
 val digest_match : actual:vodigest -> required:vodigest -> bool
 
@@ -125,6 +124,11 @@ val is_filled_opaque : Opaqueproof.opaque_handle -> safe_environment -> bool
 val repr_certificate : opaque_certificate ->
   Constr.t * Univ.ContextSet.t Opaqueproof.delayed_universes
 
+(** {5 Rewrite rules} *)
+
+(** Add a rewrite rule corresponding to the equality witnessed by the constant. *)
+val add_rewrite_rules : Label.t -> Declarations.rewrite_rules_body -> safe_environment -> safe_environment
+
 (** {5 Inductive blocks} *)
 
 (** Adding an inductive type *)
@@ -155,6 +159,7 @@ val add_constraints :
 (* val next_universe : int safe_transformer *)
 
 (** Setting the type theory flavor *)
+val set_rewrite_rules_allowed : bool -> safe_transformer0
 val set_impredicative_set : bool -> safe_transformer0
 val set_indices_matter : bool -> safe_transformer0
 val set_typing_flags : Declarations.typing_flags -> safe_transformer0
@@ -182,7 +187,7 @@ val push_named_def :
   Id.t * Entries.section_def_entry -> safe_transformer0
 
 (** Add local universes to a polymorphic section *)
-val push_section_context : Univ.UContext.t -> safe_transformer0
+val push_section_context : UVars.UContext.t -> safe_transformer0
 
 (** {6 Interactive module functions } *)
 
@@ -229,16 +234,16 @@ type compiled_library
 
 val module_of_library : compiled_library -> Declarations.module_body
 val univs_of_library : compiled_library -> Univ.ContextSet.t
+val check_flags_for_library : compiled_library -> safe_transformer0
 
 val start_library : DirPath.t -> ModPath.t safe_transformer
 
 val export :
   output_native_objects:bool ->
   safe_environment -> DirPath.t ->
-    ModPath.t * compiled_library * Nativelib.native_library
+    ModPath.t * compiled_library * Vmlibrary.compiled_library * Nativelib.native_library
 
-(* Constraints are non empty iff the file is a vi2vo *)
-val import : compiled_library -> Univ.ContextSet.t -> vodigest ->
+val import : compiled_library -> Vmlibrary.on_disk -> vodigest ->
   ModPath.t safe_transformer
 
 (** {6 Safe typing judgments } *)
@@ -267,5 +272,4 @@ val mind_of_delta_kn_senv : safe_environment -> KerName.t -> MutInd.t
 val register_inline : Constant.t -> safe_transformer0
 val register_inductive : inductive -> 'a CPrimitives.prim_ind -> safe_transformer0
 
-val set_strategy :
-  Names.Constant.t Names.tableKey -> Conv_oracle.level -> safe_transformer0
+val set_strategy : Conv_oracle.evaluable -> Conv_oracle.level -> safe_transformer0
