@@ -66,10 +66,10 @@ let raw_anomaly e = match e with
     str "Uncaught exception " ++ str (Printexc.to_string e) ++ str "."
 
 let print_backtrace e = match Exninfo.get_backtrace e with
-| None -> mt ()
+| None -> None
 | Some bt ->
   let bt = str (Exninfo.backtrace_to_string bt) in
-  fnl () ++ hov 0 bt
+  Some (hov 0 bt)
 
 let print_anomaly askreport e =
   if askreport then
@@ -96,6 +96,8 @@ let additional_error_info_handler = ref []
 let register_additional_error_info (f : Exninfo.info -> Pp.t option) =
   additional_error_info_handler := f :: !additional_error_info_handler
 
+let () = register_additional_error_info print_backtrace
+
 (** [print_gen] is a general exception printer which tries successively
     all the handlers of a list, and finally a [bottom] handler if all
     others have failed *)
@@ -109,24 +111,26 @@ let rec print_gen ~anomaly stk e =
     | Some err_msg -> err_msg
     | None -> print_gen ~anomaly stk' e
 
-let print_gen ~anomaly (e, info) =
-  let extra_msg =
+let print_extra info =
+  let pp =
     CList.map_filter (fun f -> f info) !additional_error_info_handler
-    (* Location info in the handler is ignored *)
-    |> Pp.seq
   in
+  if CList.is_empty pp then mt()
+  else fnl() ++ prlist_with_sep fnl (fun x -> x) pp
+
+let print_gen ~anomaly (e, info) =
+  let extra_msg = print_extra info in
   try
-    extra_msg ++ print_gen ~anomaly !handle_stack e
+    print_gen ~anomaly !handle_stack e ++ extra_msg
   with exn ->
     let exn, info = Exninfo.capture exn in
     (* exception in error printer *)
-    str "<in exception printer>:" ++ spc() ++ print_anomaly anomaly exn ++
-    print_backtrace info ++ fnl() ++
-    str "<original exception>:" ++ spc() ++ extra_msg ++ print_anomaly false e
+    str "<in exception printer>:" ++ spc() ++ print_anomaly anomaly exn ++ print_extra info ++ fnl() ++
+    str "<original exception>:" ++ spc() ++ print_anomaly false e ++ extra_msg
 
 (** The standard exception printer *)
 let iprint (e, info) =
-  print_gen ~anomaly:true (e,info) ++ print_backtrace info
+  print_gen ~anomaly:true (e,info)
 
 let print e =
   iprint (e, Exninfo.info e)
@@ -134,7 +138,7 @@ let print e =
 (** Same as [print], except that the "Please report" part of an anomaly
     isn't printed (used in Ltac debugging). *)
 let iprint_no_report (e, info) =
-  print_gen ~anomaly:false (e,info) ++ print_backtrace info
+  print_gen ~anomaly:false (e,info)
 let print_no_report e = iprint_no_report (e, Exninfo.info e)
 
 (** Predefined handlers **)
@@ -155,5 +159,12 @@ let noncritical = function
   | Assert_failure _ | Match_failure _ | Anomaly _
   | Control.Timeout -> false
   | Invalid_argument "equal: functional value" -> false
-  | _ -> true
+  | _ -> not (Memprof_coq.is_interrupted ())
 [@@@ocaml.warning "+52"]
+
+(* This should be in exninfo, but noncritical is here... *)
+let to_result ~f x =
+  try Ok (f x)
+  with exn when noncritical exn ->
+    let iexn = Exninfo.capture exn in
+    Error iexn
