@@ -114,6 +114,7 @@ let spc   () = Ppcmd_print_break (1,0)
 let cut   () = Ppcmd_print_break (0,0)
 let align () = Ppcmd_print_break (0,0)
 let int   n  = str (string_of_int n)
+let int64 n  = str (Int64.to_string n)
 let real  r  = str (string_of_float r)
 let bool  b  = str (string_of_bool b)
 
@@ -139,17 +140,7 @@ let hov n s = Ppcmd_box(Pp_hovbox n,s)
 (* Opening and closing of tags *)
 let tag t s = Ppcmd_tag(t,s)
 
-(* In new syntax only double quote char is escaped by repeating it *)
-let escape_string s =
-  let rec escape_at s i =
-    if i<0 then s
-    else if s.[i] == '"' then
-      let s' = String.sub s 0 i^"\""^String.sub s i (String.length s - i) in
-      escape_at s' (i-1)
-    else escape_at s (i-1) in
-  escape_at s (String.length s - 1)
-
-let qstring s = str "\"" ++ str (escape_string s) ++ str "\""
+let qstring s = str (CString.quote_coq_string s)
 let qs = qstring
 let quote s = h (str "\"" ++ s ++ str "\"")
 
@@ -164,6 +155,10 @@ let rec pr_com ft s =
       Some s2 -> Format.pp_force_newline ft (); pr_com ft s2
     | None -> ()
 
+let pr_com ft s =
+  pr_com ft s;
+  Format.pp_print_break ft 0 0
+
 let start_pfx = "start."
 let end_pfx = "end."
 
@@ -197,9 +192,9 @@ let pp_with ft pp =
     | Ppcmd_print_break(m,n)  -> pp_print_break ft m n
     | Ppcmd_force_newline     -> pp_force_newline ft ()
     | Ppcmd_comment coms      -> List.iter (pr_com ft) coms
-    | Ppcmd_tag(tag, s)       -> pp_open_tag  ft tag [@warning "-3"];
+    | Ppcmd_tag(tag, s)       -> pp_open_stag ft (String_tag tag);
                                  pp_cmd s;
-                                 pp_close_tag ft () [@warning "-3"]
+                                 pp_close_stag ft ()
   in
   pp_cmd pp
 
@@ -222,6 +217,8 @@ let pr_arg pr x = spc () ++ pr x
 let pr_non_empty_arg pr x = let pp = pr x in if ismt pp then mt () else spc () ++ pr x
 let pr_opt pr = function None -> mt () | Some x -> pr_arg pr x
 let pr_opt_no_spc pr = function None -> mt () | Some x -> pr x
+let pr_opt_default prdf pr = function None -> prdf () | Some x -> pr_arg pr x
+let pr_opt_no_spc_default prdf pr = function None -> prdf () | Some x -> pr x
 
 (** TODO: merge with CString.ordinal *)
 let pr_nth n =
