@@ -132,6 +132,8 @@ let dummy = {
 let synterp_state = ref dummy
 let interp_state = ref ([] : Summary.Interp.frozen library_segment)
 
+let library_info = ref UserWarn.empty
+
 let contents () = !interp_state
 
 let start_compilation s mp =
@@ -309,11 +311,17 @@ let library_part = function
   | GlobRef.VarRef id -> library_dp ()
   | ref -> ModPath.dp (mp_of_global ref)
 
-let discharge_proj_repr p =
+let discharge_section_proj_repr p =
   let ind = Projection.Repr.inductive p in
   let sec = section_segment_of_reference (GlobRef.IndRef ind) in
   Cooking.discharge_proj_repr sec p
 
+let discharge_proj_repr p =
+    if is_in_section (Names.GlobRef.IndRef (Names.Projection.Repr.inductive p)) then
+      discharge_section_proj_repr p
+    else
+      p
+
 (** The [LibActions] abstraction represent the set of operations on the Lib
     structure that is specific to a given stage. Two instances are defined below,
     for Synterp and Interp. *)
@@ -346,6 +354,8 @@ module type LibActions = sig
 
   val drop_objects : frozen -> frozen
 
+  val declare_info : Library_info.t -> unit
+
 end
 
 module SynterpActions : LibActions with type summary = Summary.Synterp.frozen = struct
@@ -387,6 +397,7 @@ module SynterpActions : LibActions with type summary = Summary.Synterp.frozen =
     let dir = Libnames.add_dirpath_suffix !synterp_state.path_prefix.Nametab.obj_dir id in
     let prefix = Nametab.{ obj_dir = dir; obj_mp = mp; } in
     check_mod_fresh ~is_type prefix id;
+    assert (not (sections_are_opened()));
     add_entry (OpenedModule (is_type,export,prefix,fs));
     synterp_state := { !synterp_state with path_prefix = prefix } ;
     prefix
@@ -432,6 +443,15 @@ module SynterpActions : LibActions with type summary = Summary.Synterp.frozen =
     let lib_synterp_stk = List.map (fun (node,_) -> drop_node node, []) st.lib_stk in
     { st with lib_stk = lib_synterp_stk }
 
+  let declare_info info =
+    let open UserWarn in
+    let depr = match !library_info.depr, info.depr with
+      | None, depr | depr, None -> depr
+      | Some _, Some _ ->
+         CErrors.user_err Pp.(str "Library file is already deprecated.") in
+    let warn = !library_info.warn @ info.warn in
+    library_info := { depr; warn }
+
 end
 
 module InterpActions : LibActions with type summary = Summary.Interp.frozen = struct
@@ -499,6 +519,8 @@ module InterpActions : LibActions with type summary = Summary.Interp.frozen = st
     in
     List.map (fun (node,_) -> drop_node node, []) interp_state
 
+  let declare_info _ = ()
+
 end
 
 let add_discharged_leaf obj =
@@ -563,6 +585,8 @@ module type StagedLibS = sig
 
   val drop_objects : frozen -> frozen
 
+  val declare_info : Library_info.t -> unit
+
 end
 
 (** The [StagedLib] abstraction factors out the code dealing with Lib structure
@@ -644,6 +668,8 @@ let init () = Actions.init ()
 
 let drop_objects st = Actions.drop_objects st
 
+let declare_info info = Actions.declare_info info
+
 end
 
 module Synterp : StagedLibS with type summary = Summary.Synterp.frozen = StagedLib(SynterpActions)
@@ -658,7 +684,7 @@ let end_compilation dir =
   synterp_state := { !synterp_state with comp_name = None };
   let syntax_after = Synterp.classify_segment syntax_after in
   let after = Interp.classify_segment after in
-  !synterp_state.path_prefix, after, syntax_after
+  !synterp_state.path_prefix, !library_info, after, syntax_after
 
 (** Compatibility layer *)
 let init () =
