@@ -8,9 +8,6 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-open Pp
-open Names
-
 module Dyn = Dyn.Make ()
 
 type substitutivity = Dispose | Substitute | Keep | Anticipate
@@ -53,6 +50,11 @@ let in_filter ~cat f =
 
 let simple_open ?cat f filter i o = if in_filter ~cat filter then f i o
 
+let filter_eq f1 f2 = match f1, f2 with
+  | Unfiltered, Unfiltered -> true
+  | Unfiltered, _ | _, Unfiltered -> false
+  | Filtered f1, Filtered f2 -> CString.Pred.equal f1 f2
+
 let filter_and f1 f2 = match f1, f2 with
   | Unfiltered, f | f, Unfiltered -> Some f
   | Filtered f1, Filtered f2 ->
@@ -66,6 +68,7 @@ let filter_or f1 f2 = match f1, f2 with
 
 type ('a,'b) object_declaration = {
   object_name : string;
+  object_stage : Summary.Stage.t;
   cache_function : 'b -> unit;
   load_function : int -> 'b -> unit;
   open_function : open_filter -> int -> 'b -> unit;
@@ -75,13 +78,15 @@ type ('a,'b) object_declaration = {
   rebuild_function : 'a -> 'a;
 }
 
-let default_object s = {
+let default_object ?(stage=Summary.Stage.Interp) s = {
   object_name = s;
+  object_stage = stage;
   cache_function = (fun _ -> ());
   load_function = (fun _ _ -> ());
   open_function = (fun _ _ _ -> ());
   subst_function = (fun _ ->
-    CErrors.anomaly (str "The object " ++ str s ++ str " does not know how to substitute!"));
+    CErrors.anomaly Pp.(str "The object " ++ str s
+      ++ str " does not know how to substitute!"));
   classify_function = (fun _ -> Keep);
   discharge_function = (fun _ -> None);
   rebuild_function = (fun x -> x);
@@ -116,17 +121,17 @@ type obj = Dyn.t (* persistent dynamic objects *)
 
 type algebraic_objects =
   | Objs of t list
-  | Ref of ModPath.t * Mod_subst.substitution
+  | Ref of Names.ModPath.t * Mod_subst.substitution
 
 and t =
-  | ModuleObject of Id.t * substitutive_objects
-  | ModuleTypeObject of Id.t * substitutive_objects
+  | ModuleObject of Names.Id.t * substitutive_objects
+  | ModuleTypeObject of Names.Id.t * substitutive_objects
   | IncludeObject of algebraic_objects
-  | KeepObject of Id.t * t list
-  | ExportObject of { mpl : (open_filter * ModPath.t) list }
+  | KeepObject of Names.Id.t * t list
+  | ExportObject of { mpl : (open_filter * Names.ModPath.t) list }
   | AtomicObject of obj
 
-and substitutive_objects = MBId.t list * algebraic_objects
+and substitutive_objects = Names.MBId.t list * algebraic_objects
 
 module DynMap = Dyn.Map (struct type 'a t = ('a, Nametab.object_prefix * 'a) object_declaration end)
 
@@ -139,12 +144,13 @@ let declare_object_full odecl =
   tag
 
 let make_oname Nametab.{ obj_dir; obj_mp } id =
-  Libnames.make_path obj_dir id, KerName.make obj_mp (Label.of_id id)
+  Libnames.make_path obj_dir id, Names.KerName.make obj_mp (Names.Label.of_id id)
 
 let declare_named_object_full odecl =
   let odecl =
     let oname = make_oname in
     { object_name = odecl.object_name;
+      object_stage = odecl.object_stage;
       cache_function = (fun (p, (id, o)) -> odecl.cache_function (oname p id, o));
       load_function = (fun i (p, (id, o)) -> odecl.load_function i (oname p id, o));
       open_function = (fun f i (p, (id, o)) -> odecl.open_function f i (oname p id, o));
@@ -212,48 +218,56 @@ let rebuild_object (Dyn.Dyn (tag, v)) =
   let decl = DynMap.find tag !cache_tab in
   Dyn.Dyn (tag, decl.rebuild_function v)
 
+let object_stage (Dyn.Dyn (tag, v)) =
+  let decl = DynMap.find tag !cache_tab in
+  decl.object_stage
+
 let dump = Dyn.dump
 
-let local_object_nodischarge s ~cache =
-  { (default_object s) with
+let local_object_nodischarge ?stage s ~cache =
+  { (default_object ?stage s) with
     cache_function = cache;
     classify_function = (fun _ -> Dispose);
   }
 
-let local_object s ~cache ~discharge =
-  { (local_object_nodischarge s ~cache) with
+let local_object ?stage s ~cache ~discharge =
+  { (local_object_nodischarge ?stage s ~cache) with
     discharge_function = discharge;
   }
 
-let global_object_nodischarge ?cat s ~cache ~subst =
+let global_object_nodischarge ?cat ?stage s ~cache ~subst =
   let import i o = if Int.equal i 1 then cache o in
-  { (default_object s) with
+  { (default_object ?stage s) with
     cache_function = cache;
     open_function = simple_open ?cat import;
     subst_function = (match subst with
-        | None -> fun _ -> CErrors.anomaly (str "The object " ++ str s ++ str " does not know how to substitute!")
+        | None -> fun _ ->
+            CErrors.anomaly Pp.(str "The object " ++ str s
+              ++ str " does not know how to substitute!")
         | Some subst -> subst;
       );
     classify_function =
       if Option.has_some subst then (fun _ -> Substitute) else (fun _ -> Keep);
   }
 
-let global_object ?cat s ~cache ~subst ~discharge =
+let global_object ?cat ?stage s ~cache ~subst ~discharge =
   { (global_object_nodischarge ?cat s ~cache ~subst) with
     discharge_function = discharge }
 
-let superglobal_object_nodischarge s ~cache ~subst =
-  { (default_object s) with
+let superglobal_object_nodischarge ?stage s ~cache ~subst =
+  { (default_object ?stage s) with
     load_function = (fun _ x -> cache x);
     cache_function = cache;
     subst_function = (match subst with
-        | None -> fun _ -> CErrors.anomaly (str "The object " ++ str s ++ str " does not know how to substitute!")
+        | None -> fun _ ->
+            CErrors.anomaly Pp.(str "The object " ++ str s
+              ++ str " does not know how to substitute!")
         | Some subst -> subst;
       );
     classify_function =
       if Option.has_some subst then (fun _ -> Substitute) else (fun _ -> Keep);
   }
 
-let superglobal_object s ~cache ~subst ~discharge =
-  { (superglobal_object_nodischarge s ~cache ~subst) with
+let superglobal_object ?stage s ~cache ~subst ~discharge =
+  { (superglobal_object_nodischarge ?stage s ~cache ~subst) with
     discharge_function = discharge }
