@@ -16,12 +16,43 @@ open Gramlib
 (** The parser of Coq *)
 include Grammar.GMake(CLexer.Lexer)
 
+let keyword_state = ref CLexer.empty_keyword_state
+let get_keyword_state () = !keyword_state
+let set_keyword_state s = keyword_state := s
+
+(** Not marshallable! *)
+let estate = ref EState.empty
+
+let gstate () = { GState.estate = !estate; kwstate = !keyword_state; }
+
+module Parsable = struct
+  include Parsable
+  let consume x len = consume x len !keyword_state
+end
+
+module Entry = struct
+  include Entry
+  let make na =
+    let estate', e = make na !estate in
+    estate := estate';
+    e
+  let parse e p = parse e p (gstate())
+  let of_parser na p =
+    let estate', e = of_parser na p !estate in
+    estate := estate';
+    e
+  let parse_token_stream e strm = parse_token_stream e strm (gstate())
+  let print fmt e = print fmt e !estate
+  let is_empty e = is_empty e !estate
+  let accumulate_in e = accumulate_in e !estate
+end
+
 module Lookahead =
 struct
 
   let err () = raise Stream.Failure
 
-  type t = int -> Tok.t LStream.t -> int option
+  type t = int -> CLexer.keyword_state -> (CLexer.keyword_state,Tok.t) LStream.t -> int option
 
   let rec contiguous n m strm =
     n == m ||
@@ -29,49 +60,57 @@ struct
     let (bp, _) = Loc.unloc (LStream.get_loc (n + 1) strm) in
     Int.equal ep bp && contiguous (succ n) m strm
 
-  let check_no_space m strm =
+  let check_no_space m _kwstate strm =
     let n = LStream.count strm in
     if contiguous n (n+m-1) strm then Some m else None
 
   let to_entry s (lk : t) =
-    let run strm = match lk 0 strm with None -> err () | Some _ -> () in
+    let run kwstate strm = match lk 0 kwstate strm with None -> err () | Some _ -> () in
     Entry.(of_parser s { parser_fun = run })
 
-  let (>>) (lk1 : t) lk2 n strm = match lk1 n strm with
+  let (>>) (lk1 : t) lk2 n kwstate strm = match lk1 n kwstate strm with
   | None -> None
-  | Some n -> lk2 n strm
+  | Some n -> lk2 n kwstate strm
 
-  let (<+>) (lk1 : t) lk2 n strm = match lk1 n strm with
-  | None -> lk2 n strm
+  let (<+>) (lk1 : t) lk2 n kwstate strm = match lk1 n kwstate strm with
+  | None -> lk2 n kwstate strm
   | Some n -> Some n
 
-  let lk_empty n strm = Some n
+  let lk_empty n kwstate strm = Some n
 
-  let lk_kw kw n strm = match LStream.peek_nth n strm with
+  let lk_kw kw n kwstate strm = match LStream.peek_nth kwstate n strm with
   | Tok.KEYWORD kw' | Tok.IDENT kw' -> if String.equal kw kw' then Some (n + 1) else None
   | _ -> None
 
-  let lk_kws kws n strm = match LStream.peek_nth n strm with
+  let lk_kws kws n kwstate strm = match LStream.peek_nth kwstate n strm with
   | Tok.KEYWORD kw | Tok.IDENT kw -> if List.mem_f String.equal kw kws then Some (n + 1) else None
   | _ -> None
 
-  let lk_ident n strm = match LStream.peek_nth n strm with
+  let lk_ident n kwstate strm = match LStream.peek_nth kwstate n strm with
   | Tok.IDENT _ -> Some (n + 1)
   | _ -> None
 
-  let lk_ident_except idents n strm = match LStream.peek_nth n strm with
+  let lk_name = lk_ident <+> lk_kw "_"
+
+  let lk_ident_except idents n kwstate strm = match LStream.peek_nth kwstate n strm with
   | Tok.IDENT ident when not (List.mem_f String.equal ident idents) -> Some (n + 1)
   | _ -> None
 
-  let lk_nat n strm = match LStream.peek_nth n strm with
+  let lk_nat n kwstate strm = match LStream.peek_nth kwstate n strm with
   | Tok.NUMBER p when NumTok.Unsigned.is_nat p -> Some (n + 1)
   | _ -> None
 
-  let rec lk_list lk_elem n strm =
-    ((lk_elem >> lk_list lk_elem) <+> lk_empty) n strm
+  let rec lk_list lk_elem n kwstate strm =
+    ((lk_elem >> lk_list lk_elem) <+> lk_empty) n kwstate strm
 
   let lk_ident_list = lk_list lk_ident
 
+  let lk_field n kwstate strm = match LStream.peek_nth kwstate n strm with
+    | Tok.FIELD _ -> Some (n+1)
+    | _ -> None
+
+  let lk_qualid = lk_ident >> lk_list lk_field
+
 end
 
 (** Grammar extensions *)
@@ -103,6 +142,8 @@ type ext_kind =
 
 (** The list of extensions *)
 
+let terminal s = CLexer.terminal !keyword_state s
+
 let camlp5_state = ref []
 
 let camlp5_entries = ref EntryDataMap.empty
@@ -116,9 +157,24 @@ let grammar_delete e r =
   in
   List.iter
     (fun lev ->
-      List.iter (fun pil -> safe_delete_rule e pil) (List.rev lev))
+      List.iter (fun pil -> estate := safe_delete_rule !estate e pil) (List.rev lev))
     (List.rev data)
 
+let no_add_kw = { add_kw = fun () _ -> () }
+
+(* TODO use this for ssr instead of messing with set_keyword_state
+   (needs some API design) *)
+let _safe_extend_no_kw e ext =
+  let estate', () = safe_extend no_add_kw !estate () e ext in
+  estate := estate'
+
+let add_kw = { add_kw = CLexer.add_keyword_tok }
+
+let safe_extend e ext =
+  let estate', kwstate' = safe_extend add_kw !estate !keyword_state e ext in
+  estate := estate';
+  keyword_state := kwstate'
+
 let grammar_delete_reinit e reinit d =
   grammar_delete e d;
   let a, ext = reinit in
@@ -165,14 +221,13 @@ let rec remove_grammars n =
          camlp5_state := t;
          remove_grammars (n-1)
        | ByEXTEND (name, undo,redo)::t ->
-         Feedback.msg_info Pp.(str "backtracking and redoing byextend on " ++ str name);
            undo();
            camlp5_state := t;
            remove_grammars n;
            redo();
            camlp5_state := ByEXTEND (name, undo,redo) :: !camlp5_state
        | ByEntry (tag, name, e) :: t ->
-           Unsafe.clear_entry e;
+           estate := Unsafe.clear_entry !estate e;
            camlp5_state := t;
            let EntryData.Ex entries =
              try EntryDataMap.find tag !camlp5_entries
@@ -212,8 +267,7 @@ end
 module Grammar = Register(GrammarObj)
 
 let warn_deprecated_intropattern =
-  let open CWarnings in
-  create ~name:"deprecated-intropattern-entry" ~category:"deprecated"
+  CWarnings.create ~name:"deprecated-intropattern-entry" ~category:Deprecation.Version.v8_11
   (fun () -> Pp.strbrk "Entry name intropattern has been renamed in order \
   to be consistent with the documented grammar of tactics. Use \
   \"simple_intropattern\" instead.")
@@ -228,7 +282,7 @@ let genarg_grammar x =
   Grammar.obj x
 
 let create_generic_entry2 (type a) s (etyp : a raw_abstract_argument_type) : a Entry.t =
-  let e = Entry.create s in
+  let e = Entry.make s in
   let Rawwit t = etyp in
   let () = Grammar.register0 t e in
   e
@@ -239,41 +293,41 @@ module Prim =
 
     (* Entries that can be referred via the string -> Entry.t table *)
     (* Typically for tactic or vernac extensions *)
-    let preident = Entry.create "preident"
-    let ident = Entry.create "ident"
-    let natural = Entry.create "natural"
-    let integer = Entry.create "integer"
-    let bignat = Entry.create "bignat"
-    let bigint = Entry.create "bigint"
-    let string = Entry.create "string"
-    let lstring = Entry.create "lstring"
-    let reference = Entry.create "reference"
-    let fields = Entry.create "fields"
-    let by_notation = Entry.create "by_notation"
-    let smart_global = Entry.create "smart_global"
-    let strategy_level = Entry.create "strategy_level"
+    let preident = Entry.make "preident"
+    let ident = Entry.make "ident"
+    let natural = Entry.make "natural"
+    let integer = Entry.make "integer"
+    let bignat = Entry.make "bignat"
+    let bigint = Entry.make "bigint"
+    let string = Entry.make "string"
+    let lstring = Entry.make "lstring"
+    let reference = Entry.make "reference"
+    let fields = Entry.make "fields"
+    let by_notation = Entry.make "by_notation"
+    let smart_global = Entry.make "smart_global"
+    let strategy_level = Entry.make "strategy_level"
 
     (* parsed like ident but interpreted as a term *)
-    let hyp = Entry.create "hyp"
+    let hyp = Entry.make "hyp"
     let var = hyp
 
-    let name = Entry.create "name"
-    let identref = Entry.create "identref"
-    let univ_decl = Entry.create "univ_decl"
-    let ident_decl = Entry.create "ident_decl"
-    let pattern_ident = Entry.create "pattern_ident"
+    let name = Entry.make "name"
+    let identref = Entry.make "identref"
+    let univ_decl = Entry.make "univ_decl"
+    let ident_decl = Entry.make "ident_decl"
+    let pattern_ident = Entry.make "pattern_ident"
 
     (* A synonym of ident - maybe ident will be located one day *)
-    let base_ident = Entry.create "base_ident"
+    let base_ident = Entry.make "base_ident"
 
-    let qualid = Entry.create "qualid"
-    let fullyqualid = Entry.create "fullyqualid"
-    let dirpath = Entry.create "dirpath"
+    let qualid = Entry.make "qualid"
+    let fullyqualid = Entry.make "fullyqualid"
+    let dirpath = Entry.make "dirpath"
 
-    let ne_string = Entry.create "ne_string"
-    let ne_lstring = Entry.create "ne_lstring"
+    let ne_string = Entry.make "ne_string"
+    let ne_lstring = Entry.make "ne_lstring"
 
-    let bar_cbrace = Entry.create "'|}'"
+    let bar_cbrace = Entry.make "'|}'"
 
   end
 
@@ -281,45 +335,44 @@ module Constr =
   struct
 
     (* Entries that can be referred via the string -> Entry.t table *)
-    let constr = Entry.create "constr"
-    let term = Entry.create "term"
+    let constr = Entry.make "constr"
+    let term = Entry.make "term"
     let constr_eoi = eoi_entry constr
-    let lconstr = Entry.create "lconstr"
-    let binder_constr = Entry.create "binder_constr"
-    let ident = Entry.create "ident"
-    let global = Entry.create "global"
-    let universe_name = Entry.create "universe_name"
-    let universe_level = Entry.create "universe_level"
-    let sort = Entry.create "sort"
-    let sort_family = Entry.create "sort_family"
-    let pattern = Entry.create "pattern"
-    let constr_pattern = Entry.create "constr_pattern"
-    let cpattern = Entry.create "cpattern"
-    let closed_binder = Entry.create "closed_binder"
-    let binder = Entry.create "binder"
-    let binders = Entry.create "binders"
-    let open_binders = Entry.create "open_binders"
-    let one_open_binder = Entry.create "one_open_binder"
-    let one_closed_binder = Entry.create "one_closed_binder"
-    let binders_fixannot = Entry.create "binders_fixannot"
-    let typeclass_constraint = Entry.create "typeclass_constraint"
-    let record_declaration = Entry.create "record_declaration"
-    let arg = Entry.create "arg"
-    let type_cstr = Entry.create "type_cstr"
+    let lconstr = Entry.make "lconstr"
+    let binder_constr = Entry.make "binder_constr"
+    let ident = Entry.make "ident"
+    let global = Entry.make "global"
+    let universe_name = Entry.make "universe_name"
+    let sort = Entry.make "sort"
+    let sort_family = Entry.make "sort_family"
+    let pattern = Entry.make "pattern"
+    let constr_pattern = Entry.make "constr_pattern"
+    let cpattern = Entry.make "cpattern"
+    let closed_binder = Entry.make "closed_binder"
+    let binder = Entry.make "binder"
+    let binders = Entry.make "binders"
+    let open_binders = Entry.make "open_binders"
+    let one_open_binder = Entry.make "one_open_binder"
+    let one_closed_binder = Entry.make "one_closed_binder"
+    let binders_fixannot = Entry.make "binders_fixannot"
+    let typeclass_constraint = Entry.make "typeclass_constraint"
+    let record_declaration = Entry.make "record_declaration"
+    let arg = Entry.make "arg"
+    let type_cstr = Entry.make "type_cstr"
   end
 
 module Module =
   struct
-    let module_expr = Entry.create "module_expr"
-    let module_type = Entry.create "module_type"
+    let module_expr = Entry.make "module_expr"
+    let module_type = Entry.make "module_type"
   end
 
 let epsilon_value (type s tr a) f (e : (s, tr, a) Symbol.t) =
   let r = Production.make (Rule.next Rule.stop e) (fun x _ -> f x) in
   let entry = Entry.make "epsilon" in
   let ext = Fresh (Gramlib.Gramext.First, [None, None, [r]]) in
-  let () = safe_extend entry ext in
-  try Some (parse_string entry "") with _ -> None
+  safe_extend entry ext;
+  try Some (parse_string entry "") with e when CErrors.noncritical e -> None
 
 (** Synchronized grammar extensions *)
 
@@ -436,8 +489,8 @@ type frozen_t =
   (grammar_entry * GramState.t) list *
   CLexer.keyword_state
 
-let freeze ~marshallable : frozen_t =
-  (!grammar_stack, CLexer.get_keyword_state ())
+let freeze () : frozen_t =
+  (!grammar_stack, !keyword_state)
 
 let eq_grams (g1, _) (g2, _) = match g1, g2 with
 | GramExt (_, GrammarCommand.Dyn (t1, v1)), GramExt (_, GrammarCommand.Dyn (t2, v2)) ->
@@ -471,7 +524,7 @@ let unfreeze (grams, lex) =
   let n = number_of_entries 0 undo in
   remove_grammars n;
   grammar_stack := common;
-  CLexer.set_keyword_state lex;
+  keyword_state := lex;
   List.iter extend_dyn_grammar (List.rev redo)
 
 (** No need to provide an init function : the grammar state is
@@ -487,7 +540,7 @@ let parser_summary_tag =
       Summary.init_function = Summary.nop }
 
 let with_grammar_rule_protection f x =
-  let fs = freeze ~marshallable:false in
+  let fs = freeze () in
   try let a = f x in unfreeze fs; a
   with reraise ->
     let reraise = Exninfo.capture reraise in
