@@ -18,6 +18,7 @@ open EConstr
 open Tacmach
 open Tacticals
 open Tactics
+open Induction
 open Indfun_common
 module RelDecl = Context.Rel.Declaration
 
@@ -30,14 +31,14 @@ let is_rec_info sigma scheme_info =
         (fst (decompose_prod_decls sigma (RelDecl.get_type decl)))
     in
     let free_rels_in_br = Termops.free_rels sigma new_branche in
-    let max = min + scheme_info.Tactics.npredicates in
+    let max = min + scheme_info.npredicates in
     Int.Set.exists (fun i -> i >= min && i < max) free_rels_in_br
   in
-  List.fold_left_i test_branche 1 false (List.rev scheme_info.Tactics.branches)
+  List.fold_left_i test_branche 1 false (List.rev scheme_info.branches)
 
 let choose_dest_or_ind scheme_info args =
   Proofview.tclBIND Proofview.tclEVARMAP (fun sigma ->
-      Tactics.induction_destruct (is_rec_info sigma scheme_info) false args)
+      Induction.induction_destruct (is_rec_info sigma scheme_info) false args)
 
 let functional_induction with_clean c princl pat =
   let open Proofview.Notations in
@@ -107,7 +108,7 @@ let functional_induction with_clean c princl pat =
       let sigma = project gl in
       let princ_infos = compute_elim_sig (project gl) princ_type in
       let args_as_induction_constr =
-        let c_list = if princ_infos.Tactics.farg_in_concl then [c] else [] in
+        let c_list = if princ_infos.farg_in_concl then [c] else [] in
         if List.length args + List.length c_list = 0 then
           user_err Pp.(str "Cannot recognize a valid functional scheme");
         let encoded_pat_as_patlist =
