@@ -37,33 +37,36 @@ let error msg = CErrors.user_err Pp.(str msg)
 
 type protect_flag = Eval|Prot|Rec
 
-type protection = Evd.evar_map -> EConstr.t -> GlobRef.t -> (Int.t -> protect_flag) option
+type protection = Environ.env -> Evd.evar_map -> EConstr.t -> GlobRef.t -> (Int.t -> protect_flag) option
 
 let global_head_of_constr sigma c =
   let f, args = decompose_app sigma c in
     try fst (EConstr.destRef sigma f)
     with DestKO -> CErrors.anomaly (str "global_head_of_constr.")
 
-let global_of_constr_nofail c =
-  try fst @@ Constr.destRef c
+let global_of_constr_nofail sigma c =
+  try fst @@ EConstr.destRef sigma c
   with DestKO -> GlobRef.VarRef (Id.of_string "dummy")
 
-let rec mk_clos_but f_map n t =
-  let (f, args) = Constr.decompose_appvect t in
-  match f_map (global_of_constr_nofail f) with
+let mk_atom c = CClosure.mk_atom (EConstr.Unsafe.to_constr c)
+
+let rec mk_clos_but sigma f_map n t =
+  let (f, args) = EConstr.decompose_app sigma t in
+  match f_map (global_of_constr_nofail sigma f) with
   | Some tag ->
-      let map i t = tag_arg f_map n (tag i) t in
+      let map i t = tag_arg sigma f_map n (tag i) t in
       if Array.is_empty args then map (-1) f
       else mk_red (FApp (map (-1) f, Array.mapi map args))
   | None -> mk_atom t
 
-and tag_arg f_map n tag c = match tag with
-| Eval -> mk_clos (Esubst.subs_id n) c
+and tag_arg sigma f_map n tag c = match tag with
+| Eval -> mk_clos (Esubst.subs_id n, UVars.Instance.empty) (EConstr.Unsafe.to_constr c)
 | Prot -> mk_atom c
-| Rec -> mk_clos_but f_map n c
+| Rec -> mk_clos_but sigma f_map n c
 
-let interp_map l t =
-  try Some(List.assoc_f GlobRef.equal t l) with Not_found -> None
+let interp_map env l t =
+  let eq g1 g2 = QGlobRef.equal env g1 g2 in
+  try Some (List.assoc_f eq t l) with Not_found -> None
 
 let protect_maps : protection String.Map.t ref = ref String.Map.empty
 let add_map s m = protect_maps := String.Map.add s m !protect_maps
@@ -72,16 +75,15 @@ let lookup_map map =
   with Not_found ->
     CErrors.user_err (str "Map " ++ qs map ++ str "not found.")
 
-let protect_red map env sigma c0 =
-  let c = EConstr.Unsafe.to_constr c0 in
+let protect_red map env sigma c =
   let tab = create_tab () in
-  let infos = Evarutil.create_clos_infos env sigma all in
-  let map = lookup_map map sigma c0 in
-  let rec eval n c = match Constr.kind c with
-  | Prod (na, t, u) -> Constr.mkProd (na, eval n t, eval (n + 1) u)
-  | _ -> kl infos tab (mk_clos_but map n c)
+  let infos = Evarutil.create_clos_infos env sigma RedFlags.all in
+  let map = lookup_map map env sigma c in
+  let rec eval n c = match EConstr.kind sigma c with
+  | Prod (na, t, u) -> EConstr.mkProd (na, eval n t, eval (n + 1) u)
+  | _ -> EConstr.of_constr @@ norm_val infos tab (mk_clos_but sigma map n c)
   in
-  EConstr.of_constr (eval 0 c)
+  eval 0 c
 
 let protect_tac map =
   Tactics.reduct_option ~check:false (protect_red map,DEFAULTcast) None
@@ -108,12 +110,12 @@ let closed_term args _ = match args with
   let l = List.map (fun c -> Value.cast (Genarg.topwit Stdarg.wit_ref) c) (Option.get (Value.to_list l)) in
   Proofview.tclEVARMAP >>= fun sigma ->
   let cs = List.fold_right GlobRef.Set_env.add l GlobRef.Set_env.empty in
-  if closed_under sigma cs t then Proofview.tclUNIT () else Tacticals.tclFAIL 0 (mt())
+  if closed_under sigma cs t then Proofview.tclUNIT () else Tacticals.tclFAIL (mt())
 | _ -> assert false
 
 let closed_term_ast =
   let tacname = {
-    mltac_plugin = "ring_plugin";
+    mltac_plugin = "coq-core.plugins.ring";
     mltac_tactic = "closed_term";
   } in
   let () = Tacenv.register_ml_tactic tacname [|closed_term|] in
@@ -144,13 +146,15 @@ let ic_unsafe env sigma c = (*FIXME remove *)
 let decl_constant name univs c =
   let open Constr in
   let vars = CVars.universes_of_constr c in
-  let univs = UState.restrict_universe_context ~lbound:(Global.universes_lbound ()) univs vars in
-  let () = DeclareUctx.declare_universe_context ~poly:false univs in
+  let univs = UState.restrict_universe_context univs vars in
+  let () = Global.push_context_set ~strict:true univs in
   let types = (Typeops.infer (Global.env ()) c).uj_type in
   let univs = UState.Monomorphic_entry Univ.ContextSet.empty, UnivNames.empty_binders in
-  mkConst(declare_constant ~name
-            ~kind:Decls.(IsProof Lemma)
-            (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
+  (* UnsafeMonomorphic: we always do poly:false *)
+  UnsafeMonomorphic.mkConst
+    (declare_constant ~name
+       ~kind:Decls.(IsProof Lemma)
+       (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
 
 let decl_constant na suff univs c =
   let na = Namegen.next_global_ident_away (Nameops.add_suffix na suff) Id.Set.empty in
@@ -168,7 +172,7 @@ let tactic_res = ref [||]
 
 let get_res =
   let open Tacexpr in
-  let name = { mltac_plugin = "ring_plugin"; mltac_tactic = "get_res"; } in
+  let name = { mltac_plugin = "coq-core.plugins.ring"; mltac_tactic = "get_res"; } in
   let entry = { mltac_name = name; mltac_index = 0 } in
   let tac args ist =
     let n = Tacinterp.Value.cast (Genarg.topwit Stdarg.wit_int) (List.hd args) in
@@ -288,14 +292,14 @@ let coq_mkhypo = my_reference "mkhypo"
 let coq_hypo = my_reference "hypo"
 
 (* Equality: do not evaluate but make recursive call on both sides *)
-let map_with_eq arg_map sigma c =
+let map_with_eq arg_map env sigma c =
   let (req,_,_) = dest_rel sigma c in
-  interp_map
+  interp_map env
     ((global_head_of_constr sigma req,(function -1->Prot|_->Rec))::
     List.map (fun (c,map) -> (Lazy.force c,map)) arg_map)
 
-let map_without_eq arg_map _ _ =
-  interp_map (List.map (fun (c,map) -> (Lazy.force c,map)) arg_map)
+let map_without_eq arg_map env _ _ =
+  interp_map env (List.map (fun (c,map) -> (Lazy.force c,map)) arg_map)
 
 let _ = add_map "ring"
   (map_with_eq
@@ -344,7 +348,7 @@ let find_ring_structure env sigma l =
               (str"Arguments of ring_simplify do not have all the same type.")
         in
         List.iter check cl';
-        (try ring_for_carrier (EConstr.to_constr sigma ty)
+        (try ring_for_carrier (EConstr.to_constr ~abort_on_undefined_evars:false sigma ty)
         with Not_found ->
           CErrors.user_err
             (str"Cannot find a declared ring structure over"++
@@ -396,9 +400,8 @@ let subst_th (subst,th) =
 
 
 let theory_to_obj : ring_info -> obj =
-  let cache_th (_, th) = add_entry th in
   declare_object @@ global_object_nodischarge "tactic-new-ring-theory"
-    ~cache:cache_th
+    ~cache:add_entry
     ~subst:(Some subst_th)
 
 let setoid_of_relation env sigma a r =
@@ -431,18 +434,18 @@ let ring_equality env sigma (r,add,mul,opp,req) =
         let sigma, setoid = setoid_of_relation env sigma r req in
         let signature = [Some (r,Some req);Some (r,Some req)],Some(r,Some req) in
         let add_m, add_m_lem =
-          try Rewrite.default_morphism signature add
+          try Rewrite.Internal.default_morphism env sigma signature add
           with Not_found ->
             CErrors.user_err (str "Ring addition " ++ pr_econstr_env env sigma add ++ str " should be declared as a morphism.") in
         let mul_m, mul_m_lem =
-          try Rewrite.default_morphism signature mul
+          try Rewrite.Internal.default_morphism env sigma signature mul
           with Not_found ->
             CErrors.user_err (str "Ring multiplication " ++ pr_econstr_env env sigma mul ++ str " should be declared as a morphism.") in
         let op_morph =
           match opp with
             | Some opp ->
                 (let opp_m,opp_m_lem =
-                  try Rewrite.default_morphism ([Some(r,Some req)],Some(r,Some req)) opp
+                  try Rewrite.Internal.default_morphism env sigma ([Some(r,Some req)],Some(r,Some req)) opp
                   with Not_found ->
                     CErrors.user_err (str "Ring opposite " ++ pr_econstr_env env sigma opp ++ str " should be declared as a morphism.") in
                 let op_morph =
@@ -515,8 +518,7 @@ let make_hyp_list env sigma lH =
         (plapp sigma coq_nil [|carrier|])
   in
   let sigma, l' = Typing.solve_evars env sigma l in
-  let l' = EConstr.Unsafe.to_constr l' in
-  sigma, Evarutil.nf_evars_universes sigma l'
+  sigma, l'
 
 let interp_power env sigma pow =
   let sigma, carrier = Evd.fresh_global env sigma (Lazy.force coq_hypo) in
@@ -586,7 +588,7 @@ let add_theory0 env sigma name rth eqth morphth cst_tac (pre,post) power sign di
   let req = EConstr.to_constr sigma req in
   let sth = EConstr.to_constr sigma sth in
   let _ =
-    Lib.add_anonymous_leaf
+    Lib.add_leaf
       (theory_to_obj
         { ring_name = name;
           ring_carrier = r;
@@ -684,7 +686,7 @@ let ring_lookup (f : Value.t) lH rl t =
     let sigma, l = make_term_list env sigma (EConstr.of_constr e.ring_carrier) rl in
     let rl = Value.of_constr l in
     let sigma, l = make_hyp_list env sigma lH in
-    let lH = carg l in
+    let lH = Value.of_constr l in
     let ring = ltac_ring_structure e in
     Proofview.tclTHEN (Proofview.Unsafe.tclEVARS sigma) (Value.apply f (ring@[lH;rl]))
   end
@@ -748,18 +750,18 @@ let dest_field env sigma th_spec =
   let th_typ = Retyping.get_type_of env sigma th_spec in
   match EConstr.kind sigma th_typ with
     | App(f,[|r;zero;one;add;mul;sub;opp;div;inv;req|])
-        when isRefX sigma (Lazy.force afield_theory) f ->
+        when isRefX env sigma (Lazy.force afield_theory) f ->
         let sigma, rth = plapp sigma af_ar
           [|r;zero;one;add;mul;sub;opp;div;inv;req;th_spec|] in
         (None,r,zero,one,add,mul,Some sub,Some opp,div,inv,req,rth)
     | App(f,[|r;zero;one;add;mul;sub;opp;div;inv;req|])
-        when isRefX sigma (Lazy.force field_theory) f ->
+        when isRefX env sigma (Lazy.force field_theory) f ->
         let sigma, rth =
           plapp sigma f_r
             [|r;zero;one;add;mul;sub;opp;div;inv;req;th_spec|] in
         (Some false,r,zero,one,add,mul,Some sub,Some opp,div,inv,req,rth)
     | App(f,[|r;zero;one;add;mul;div;inv;req|])
-        when isRefX sigma (Lazy.force sfield_theory) f ->
+        when isRefX env sigma (Lazy.force sfield_theory) f ->
         let sigma, rth = plapp sigma sf_sr
           [|r;zero;one;add;mul;div;inv;req;th_spec|] in
         (Some true,r,zero,one,add,mul,None,None,div,inv,req,rth)
@@ -841,9 +843,8 @@ let subst_th (subst,th) =
       field_post_tac = posttac' }
 
 let ftheory_to_obj : field_info -> obj =
-  let cache_th (_, th) = add_field_entry th in
   declare_object @@ global_object_nodischarge "tactic-new-field-theory"
-    ~cache:cache_th
+    ~cache:add_field_entry
     ~subst:(Some subst_th)
 
 let field_equality env sigma r inv req =
@@ -856,7 +857,7 @@ let field_equality env sigma r inv req =
         let _setoid = setoid_of_relation env sigma r req in
         let signature = [Some (r,Some req)],Some(r,Some req) in
         let inv_m, inv_m_lem =
-          try Rewrite.default_morphism signature inv
+          try Rewrite.Internal.default_morphism env sigma signature inv
           with Not_found ->
             error "field inverse should be declared as a morphism" in
           inv_m_lem
@@ -909,7 +910,7 @@ let add_field_theory0 env sigma name fth eqth morphth cst_tac inj (pre,post) pow
   let r = EConstr.to_constr sigma r in
   let req = EConstr.to_constr sigma req in
   let _ =
-    Lib.add_anonymous_leaf
+    Lib.add_leaf
       (ftheory_to_obj
         { field_name = name;
           field_carrier = r;
@@ -977,7 +978,7 @@ let field_lookup (f : Value.t) lH rl t =
     let sigma, c = make_term_list env sigma (EConstr.of_constr e.field_carrier) rl in
     let rl = Value.of_constr c in
     let sigma, l = make_hyp_list env sigma lH in
-    let lH = carg l in
+    let lH = Value.of_constr l in
     let field = ltac_field_structure e in
     Proofview.tclTHEN (Proofview.Unsafe.tclEVARS sigma) (Value.apply f (field@[lH;rl]))
   end
