@@ -180,6 +180,4 @@ type ssrapplyarg = ssrterm list * (ssrterm ssragens * ssripats)
 (* OOP : these are general shortcuts *)
 type gist = Tacintern.glob_sign
 type ist = Tacinterp.interp_sign
-type goal = Goal.goal
-type 'a sigma = 'a Evd.sigma
-type v82tac = Tacmach.Old.tactic
+type goal = Evar.t
