@@ -69,10 +69,26 @@ val unify : ?flags:unify_flags -> ?with_ho:bool ->
 val solve_unif_constraints_with_heuristics :
   env -> ?flags:unify_flags -> ?with_ho:bool -> evar_map -> evar_map
 
-(** Check all pending unification problems are solved and raise a
-    PretypeError otherwise *)
+(** Check all pending unification problems relative to a set of evars
+    are solved and raise a PretypeError otherwise *)
 
-val check_problems_are_solved : env -> evar_map -> unit
+val check_problems_are_solved : ?evars:Evar.Set.t -> env -> evar_map -> unit
+
+(** Hook for the canonical structure resolution *)
+
+type hook = Environ.env -> Evd.evar_map -> ((Names.Constant.t * EConstr.EInstance.t) * EConstr.t list * EConstr.t) -> (EConstr.t * EConstr.t list) -> (Evd.evar_map * Structures.CanonicalSolution.t) option
+
+val all_hooks : hook CString.Map.t ref
+
+val register_hook : name:CString.Map.key -> ?override:bool -> hook -> unit
+
+val active_hooks : string list ref
+
+val deactivate_hook : name:Util.String.t -> unit
+
+val activate_hook : name:CString.Map.key -> unit
+
+val apply_hooks : hook
 
 (** Check if a canonical structure is applicable *)
 
@@ -94,8 +110,7 @@ val compare_heads : env -> evar_map ->
     matching, using typing to select occurrences *)
 
 type occurrence_match_test =
-  env -> evar_map -> constr -> (* Used to precompute the local tests *)
-  env -> evar_map -> int -> constr -> constr -> bool * evar_map
+  env -> evar_map -> constr -> constr -> bool * evar_map
 
 (** When given the choice of abstracting an occurrence or leaving it,
     force abstration. *)
