@@ -76,21 +76,24 @@ open Goptions
 
 let () =
   declare_bool_option
-  { optdepr  = false;
+  { optstage = Summary.Stage.Interp;
+    optdepr  = None;
     optkey   = ["Transparent";"Obligations"];
     optread  = get_proofs_transparency;
     optwrite = set_proofs_transparency; }
 
 let () =
   declare_bool_option
-  { optdepr  = false;
+  { optstage = Summary.Stage.Interp;
+    optdepr  = None;
     optkey   = ["Program";"Cases"];
     optread  = (fun () -> !program_cases);
     optwrite = (:=) program_cases }
 
 let () =
   declare_bool_option
-  { optdepr  = false;
+  { optstage = Summary.Stage.Interp;
+    optdepr  = None;
     optkey   = ["Program";"Generalized";"Coercion"];
     optread  = (fun () -> !program_generalized_coercion);
     optwrite = (:=) program_generalized_coercion }
