@@ -46,14 +46,13 @@ val print_goal_tag_opt_name : string list
     [~impargs:some_list_of_binding_kind] indicates the implicit arguments
     of the external quatification. Only used for printing types (not
     terms), and at toplevel (only "l" versions). It defaults to [None].
-
-    [~lax:true] is for debugging purpose. It defaults to [~lax:false]. *)
+*)
 
 
-val pr_constr_env          : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> constr -> Pp.t
-val pr_lconstr_env         : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> constr -> Pp.t
+val pr_constr_env          : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> constr -> Pp.t
+val pr_lconstr_env         : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> constr -> Pp.t
 
-val pr_constr_n_env        : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> Constrexpr.entry_relative_level -> constr -> Pp.t
+val pr_constr_n_env        : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> Constrexpr.entry_relative_level -> constr -> Pp.t
 
 (** Same, but resilient to [Nametab] errors. Prints fully-qualified
     names when [shortest_qualid_of_global] has failed. Prints "??"
@@ -62,16 +61,16 @@ val pr_constr_n_env        : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> en
 val safe_pr_constr_env  : env -> evar_map -> constr -> Pp.t
 val safe_pr_lconstr_env : env -> evar_map -> constr -> Pp.t
 
-val pr_econstr_env      : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> EConstr.t -> Pp.t
-val pr_leconstr_env     : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> EConstr.t -> Pp.t
+val pr_econstr_env      : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> EConstr.t -> Pp.t
+val pr_leconstr_env     : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> EConstr.t -> Pp.t
 
-val pr_econstr_n_env    : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> Constrexpr.entry_relative_level -> EConstr.t -> Pp.t
+val pr_econstr_n_env    : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> Constrexpr.entry_relative_level -> EConstr.t -> Pp.t
 
-val pr_etype_env        : ?lax:bool -> ?goal_concl_style:bool -> env -> evar_map -> EConstr.types -> Pp.t
-val pr_letype_env       : ?lax:bool -> ?goal_concl_style:bool -> env -> evar_map -> ?impargs:Glob_term.binding_kind list -> EConstr.types -> Pp.t
+val pr_etype_env        : ?goal_concl_style:bool -> env -> evar_map -> EConstr.types -> Pp.t
+val pr_letype_env       : ?goal_concl_style:bool -> env -> evar_map -> ?impargs:Glob_term.binding_kind list -> EConstr.types -> Pp.t
 
-val pr_open_constr_env  : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> open_constr -> Pp.t
-val pr_open_lconstr_env : ?lax:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> open_constr -> Pp.t
+val pr_open_constr_env  : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> open_constr -> Pp.t
+val pr_open_lconstr_env : ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> open_constr -> Pp.t
 
 val pr_constr_under_binders_env  : env -> evar_map -> constr_under_binders -> Pp.t
 
@@ -96,14 +95,13 @@ val pr_lconstr_under_binders_env : env -> evar_map -> constr_under_binders -> Pp
     names of global definitions are printed qualified and only names
     of goal/section variables and rel names that do _not_ occur in the
     scope of the binder to be printed are avoided.
+*)
 
-    [~lax:true] is for debugging purpose. *)
+val pr_ltype_env           : ?goal_concl_style:bool -> env -> evar_map -> ?impargs:Glob_term.binding_kind list -> types -> Pp.t
+val pr_type_env            : ?goal_concl_style:bool -> env -> evar_map -> types -> Pp.t
 
-val pr_ltype_env           : ?lax:bool -> ?goal_concl_style:bool -> env -> evar_map -> ?impargs:Glob_term.binding_kind list -> types -> Pp.t
-val pr_type_env            : ?lax:bool -> ?goal_concl_style:bool -> env -> evar_map -> types -> Pp.t
-
-val pr_closed_glob_n_env   : ?lax:bool -> ?goal_concl_style:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> Constrexpr.entry_relative_level -> closed_glob_constr -> Pp.t
-val pr_closed_glob_env     : ?lax:bool -> ?goal_concl_style:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> closed_glob_constr -> Pp.t
+val pr_closed_glob_n_env   : ?goal_concl_style:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> Constrexpr.entry_relative_level -> closed_glob_constr -> Pp.t
+val pr_closed_glob_env     : ?goal_concl_style:bool -> ?inctx:bool -> ?scope:scope_name -> env -> evar_map -> closed_glob_constr -> Pp.t
 
 val pr_ljudge_env          : env -> evar_map -> EConstr.unsafe_judgment -> Pp.t * Pp.t
 
@@ -111,9 +109,9 @@ val pr_lglob_constr_env    : env -> evar_map -> 'a glob_constr_g -> Pp.t
 
 val pr_glob_constr_env     : env -> evar_map -> 'a glob_constr_g -> Pp.t
 
-val pr_lconstr_pattern_env : env -> evar_map -> constr_pattern -> Pp.t
+val pr_lconstr_pattern_env : env -> evar_map -> _ constr_pattern_r -> Pp.t
 
-val pr_constr_pattern_env  : env -> evar_map -> constr_pattern -> Pp.t
+val pr_constr_pattern_env  : env -> evar_map -> _ constr_pattern_r -> Pp.t
 
 val pr_cases_pattern       : cases_pattern -> Pp.t
 
@@ -121,15 +119,15 @@ val pr_sort                : evar_map -> Sorts.t -> Pp.t
 
 (** Universe constraints *)
 
-val pr_universe_instance   : evar_map -> Univ.Instance.t -> Pp.t
-val pr_universe_instance_constraints : evar_map -> Univ.Instance.t -> Univ.Constraints.t -> Pp.t
-val pr_universe_ctx        : evar_map -> ?variance:Univ.Variance.t array ->
-  Univ.UContext.t -> Pp.t
-val pr_abstract_universe_ctx : evar_map -> ?variance:Univ.Variance.t array ->
-  ?priv:Univ.ContextSet.t -> Univ.AbstractContext.t -> Pp.t
+val pr_universe_instance   : evar_map -> UVars.Instance.t -> Pp.t
+val pr_universe_instance_constraints : evar_map -> UVars.Instance.t -> Univ.Constraints.t -> Pp.t
+val pr_universe_ctx        : evar_map -> ?variance:UVars.Variance.t array ->
+  UVars.UContext.t -> Pp.t
+val pr_abstract_universe_ctx : evar_map -> ?variance:UVars.Variance.t array ->
+  ?priv:Univ.ContextSet.t -> UVars.AbstractContext.t -> Pp.t
 val pr_universe_ctx_set    : evar_map -> Univ.ContextSet.t -> Pp.t
 val pr_universes  : evar_map ->
-  ?variance:Univ.Variance.t array -> ?priv:Univ.ContextSet.t ->
+  ?variance:UVars.Variance.t array -> ?priv:Univ.ContextSet.t ->
   Declarations.universes -> Pp.t
 
 (** [universe_binders_with_opt_names ref l]
@@ -141,8 +139,8 @@ val pr_universes  : evar_map ->
     Otherwise return the bound universe names registered for [ref].
 
     Inefficient on large contexts due to name generation. *)
-val universe_binders_with_opt_names : Univ.AbstractContext.t ->
-  UnivNames.univ_name_list option -> UnivNames.universe_binders
+val universe_binders_with_opt_names : UVars.AbstractContext.t ->
+  UnivNames.full_name_list option -> UnivNames.universe_binders * UnivNames.rev_binders
 
 (** Printing global references using names as short as possible *)
 
@@ -154,12 +152,14 @@ val pr_existential_key     : env -> evar_map -> Evar.t -> Pp.t
 val pr_existential         : env -> evar_map -> existential -> Pp.t
 val pr_constructor         : env -> constructor -> Pp.t
 val pr_inductive           : env -> inductive -> Pp.t
-val pr_evaluable_reference : Tacred.evaluable_global_reference -> Pp.t
+val pr_evaluable_reference : Evaluable.t -> Pp.t
 
 val pr_pconstant : env -> evar_map -> pconstant -> Pp.t
 val pr_pinductive : env -> evar_map -> pinductive -> Pp.t
 val pr_pconstructor : env -> evar_map -> pconstructor -> Pp.t
 
+val pr_notation_interpretation_env : env -> evar_map -> glob_constr -> Pp.t
+val pr_notation_interpretation : glob_constr -> Pp.t
 
 (** Contexts *)
 
@@ -174,6 +174,10 @@ val pr_named_decl          : env -> evar_map -> Constr.named_declaration -> Pp.t
 val pr_compacted_decl      : env -> evar_map -> Constr.compacted_declaration -> Pp.t
 val pr_rel_decl            : env -> evar_map -> Constr.rel_declaration -> Pp.t
 
+val pr_enamed_decl          : env -> evar_map -> EConstr.named_declaration -> Pp.t
+val pr_ecompacted_decl      : env -> evar_map -> EConstr.compacted_declaration -> Pp.t
+val pr_erel_decl            : env -> evar_map -> EConstr.rel_declaration -> Pp.t
+
 val pr_named_context       : env -> evar_map -> Constr.named_context -> Pp.t
 val pr_named_context_of    : env -> evar_map -> Pp.t
 val pr_rel_context         : env -> evar_map -> Constr.rel_context -> Pp.t
@@ -185,6 +189,7 @@ val pr_context_of          : env -> evar_map -> Pp.t
 val pr_predicate           : ('a -> Pp.t) -> (bool * 'a list) -> Pp.t
 val pr_cpred               : Cpred.t -> Pp.t
 val pr_idpred              : Id.Pred.t -> Pp.t
+val pr_prpred              : PRpred.t -> Pp.t
 val pr_transparent_state   : TransparentState.t -> Pp.t
 
 (** Proofs, these functions obey [Hyps Limit] and [Compact contexts]. *)
@@ -196,13 +201,11 @@ val pr_transparent_state   : TransparentState.t -> Pp.t
 *)
 val pr_open_subgoals       : ?quiet:bool -> ?diffs:Proof.t option -> Proof.t -> Pp.t
 val pr_nth_open_subgoal    : proof:Proof.t -> int -> Pp.t
-val pr_evar                : evar_map -> (Evar.t * evar_info) -> Pp.t
-val pr_evars_int           : evar_map -> shelf:Evar.t list -> given_up:Evar.t list -> int -> evar_info Evar.Map.t -> Pp.t
+val pr_evar                : evar_map -> (Evar.t * undefined evar_info) -> Pp.t
+val pr_evars_int           : evar_map -> shelf:Evar.t list -> given_up:Evar.t list -> int -> undefined evar_info Evar.Map.t -> Pp.t
 val pr_ne_evar_set         : Pp.t -> Pp.t -> evar_map ->
   Evar.Set.t -> Pp.t
 
-val print_dependent_evars : Evar.t option -> evar_map -> Evar.t list -> Pp.t
-
 (** Declarations for the "Print Assumption" command *)
 type axiom =
   | Constant of Constant.t (* An axiom or a constant. *)
