@@ -29,9 +29,9 @@ type diffOpt = DiffOff | DiffOn | DiffRemoved
 
 val string_to_diffs : string -> diffOpt
 
-open Evd
-open Environ
-open Constr
+type goal
+
+val make_goal : Environ.env -> Evd.evar_map -> Evar.t -> goal
 
 (** Computes the diff between the goals of two Proofs and returns
 the highlighted lists of hypotheses and conclusions.
@@ -42,31 +42,22 @@ raise Diff_Failure under some "impossible" conditions.
 
 If you want to make your call especially bulletproof, catch these
 exceptions, print a user-visible message, then recall this routine with
-the first argument set to None, which will skip the diff.
+the second argument ('og_s') set to None, which will skip the diff.
+
+The 'short' flag skips computing the diffs of the hypotheses, which allows the
+Subgoals XML command to fetch just the conclusion of a goal.  This is useful,
+for example, when an IDE only needs to display the number of admitted goals or
+preview the next unsolved goal.
 *)
-val diff_goal_ide : (Goal.goal * Evd.evar_map) option -> Goal.goal -> Evd.evar_map -> Pp.t list * Pp.t
-
-(** Computes the diff between two goals
-
-If the strings used to display the goal are not lexable (this is believed
-unlikely), this routine will generate a Diff_Failure.  This routine may also
-raise Diff_Failure under some "impossible" conditions.
-
-If you want to make your call especially bulletproof, catch these
-exceptions, print a user-visible message, then recall this routine with
-the first argument set to None, which will skip the diff.
-*)
-val diff_goal : ?og_s:(Goal.goal * Evd.evar_map) -> Goal.goal -> Evd.evar_map -> Pp.t
+val diff_goal : ?short:bool -> ?og_s:goal -> goal -> Pp.t list * Pp.t
 
 (** Convert a string to a list of token strings using the lexer *)
 val tokenize_string : string -> string list
 
-val pr_letype_env          : ?lax:bool -> ?goal_concl_style:bool -> Environ.env -> Evd.evar_map -> ?impargs:Glob_term.binding_kind list -> EConstr.types -> Pp.t
-val pr_leconstr_env        : ?lax:bool -> ?inctx:bool -> ?scope:Notation_term.scope_name -> Environ.env -> Evd.evar_map -> EConstr.constr -> Pp.t
-val pr_lconstr_env         : ?lax:bool -> ?inctx:bool -> ?scope:Notation_term.scope_name -> env -> evar_map -> constr -> Pp.t
-
 (** Computes diffs for a single conclusion *)
-val diff_concl : ?og_s:(Goal.goal * Evd.evar_map) -> Evd.evar_map -> Goal.goal -> Pp.t
+val diff_concl : ?og_s:goal -> goal -> Pp.t
+
+type goal_map
 
 (** Generates a map from [np] to [op] that maps changed goals to their prior
 forms.  The map doesn't include entries for unchanged goals; unchanged goals
@@ -74,7 +65,9 @@ will have the same goal id in both versions.
 
 [op] and [np] must be from the same proof document and [op] must be for a state
 before [np]. *)
-val make_goal_map : Proof.t option -> Proof.t -> Goal.goal Evar.Map.t
+val make_goal_map : Proof.t -> Proof.t -> goal_map
+
+val map_goal : Evar.t -> goal_map -> goal option
 
 (* Exposed for unit test, don't use these otherwise *)
 (* output channel for the test log file *)
