@@ -16,7 +16,6 @@ open Constr
 open Termops
 open EConstr
 open Tactics
-open Clenv
 open Typeclasses
 open Evd
 open Locus
@@ -31,10 +30,11 @@ let typeclasses_db = "typeclass_instances"
 (** Options handling *)
 
 let typeclasses_depth_opt_name = ["Typeclasses";"Depth"]
-let get_typeclasses_depth =
+let { Goptions.get = get_typeclasses_depth } =
   Goptions.declare_intopt_option_and_ref
-    ~depr:false
     ~key:typeclasses_depth_opt_name
+    ~value:None
+    ()
 
 let set_typeclasses_depth =
   Goptions.set_int_option_value typeclasses_depth_opt_name
@@ -43,34 +43,24 @@ let set_typeclasses_depth =
     useless introductions. This is no longer useful since we have eta, but is
     here for compatibility purposes. Another compatibility issues is that the
     cost (in terms of search depth) can differ. *)
-let get_typeclasses_limit_intros =
+let { Goptions.get = get_typeclasses_limit_intros } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Typeclasses";"Limit";"Intros"]
     ~value:true
+    ()
 
-let get_typeclasses_dependency_order =
+let { Goptions.get = get_typeclasses_dependency_order } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Typeclasses";"Dependency";"Order"]
     ~value:false
+    ()
 
 let iterative_deepening_opt_name = ["Typeclasses";"Iterative";"Deepening"]
-let get_typeclasses_iterative_deepening =
+let { Goptions.get = get_typeclasses_iterative_deepening } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:iterative_deepening_opt_name
     ~value:false
-
-(** [typeclasses_filtered_unif] governs the unification algorithm used by type
-    classes. If enabled, a new algorithm based on pattern filtering and refine
-    will be used. When disabled, the previous algorithm based on apply will be
-    used. *)
-let get_typeclasses_filtered_unification =
-  Goptions.declare_bool_option_and_ref
-    ~depr:true
-    ~key:["Typeclasses";"Filtered";"Unification"]
-    ~value:false
+    ()
 
 module Debug : sig
   val ppdebug : int -> (unit -> Pp.t) -> unit
@@ -93,7 +83,8 @@ end = struct
   let () =
     let open Goptions in
     declare_bool_option
-      { optdepr  = false;
+      { optstage = Summary.Stage.Interp;
+        optdepr  = None;
         optkey   = ["Typeclasses";"Debug"];
         optread  = get_typeclasses_debug;
         optwrite = set_typeclasses_debug; }
@@ -101,7 +92,8 @@ end = struct
   let () =
     let open Goptions in
     declare_int_option
-      { optdepr  = false;
+      { optstage = Summary.Stage.Interp;
+        optdepr  = None;
         optkey   = ["Typeclasses";"Debug";"Verbosity"];
         optread  = get_typeclasses_verbose;
         optwrite = set_typeclasses_verbose; }
@@ -121,7 +113,7 @@ let set_typeclasses_strategy = function
   | Bfs -> Goptions.set_bool_option_value iterative_deepening_opt_name true
 
 let pr_ev evs ev =
-  let evi = Evd.find evs ev in
+  let evi = Evd.find_undefined evs ev in
   let env = Evd.evar_filtered_env (Global.env ()) evi in
   Printer.pr_econstr_env env evs (Evd.evar_concl evi)
 
@@ -165,7 +157,7 @@ let e_give_exact flags h =
   let sigma, c = Hints.fresh_hint env sigma h in
   let (sigma, t1) = Typing.type_of (pf_env gl) sigma c in
   Proofview.Unsafe.tclEVARS sigma <*>
-  Clenv.unify ~flags t1 <*> exact_no_check c
+  Clenv.unify ~flags ~cv_pb:CUMUL t1 <*> exact_no_check c
   end
 
 let unify_resolve ~with_evars flags h diff = match diff with
@@ -177,40 +169,10 @@ let unify_resolve ~with_evars flags h diff = match diff with
   let env = Proofview.Goal.env gl in
   let sigma = Tacmach.project gl in
   let sigma, c = Hints.fresh_hint env sigma h in
-  let clenv = mk_clenv_from_n env sigma diff (c, ty) in
+  let clenv = Clenv.mk_clenv_from_n env sigma diff (c, ty) in
   Clenv.res_pf ~with_evars ~with_classes:false ~flags clenv
   end
 
-(** Application of a lemma using [refine] instead of the old [w_unify] *)
-let unify_resolve_refine flags h diff =
-  let len = match diff with None -> None | Some (diff, _) -> Some diff in
-  Proofview.Goal.enter begin fun gls ->
-  let open Clenv in
-  let env = Proofview.Goal.env gls in
-  let concl = Proofview.Goal.concl gls in
-  Refine.refine ~typecheck:false begin fun sigma ->
-    let sigma, term = Hints.fresh_hint env sigma h in
-    let ty = Retyping.get_type_of env sigma term in
-    let sigma, cl = Clenv.make_evar_clause env sigma ?len ty in
-    let term = applist (term, List.map (fun x -> x.hole_evar) cl.cl_holes) in
-    let flags = Evarconv.default_flags_of flags.core_unify_flags.modulo_delta in
-    let sigma = Evarconv.unify_leq_delay ~flags env sigma cl.cl_concl concl in
-    (sigma, term)
-    end
-  end
-
-let unify_resolve_refine flags h diff =
-  Proofview.tclORELSE
-    (unify_resolve_refine flags h diff)
-    (fun (exn,info) ->
-      match exn with
-      | Evarconv.UnableToUnify _ ->
-        Tacticals.tclZEROMSG ~info (str "Unable to unify")
-      | e when CErrors.noncritical e ->
-        Tacticals.tclZEROMSG ~info (str "Unexpected error")
-      | _ ->
-        Exninfo.iraise (exn,info))
-
 (** Dealing with goals of the form A -> B and hints of the form
   C -> A -> B.
 *)
@@ -230,21 +192,6 @@ let with_prods nprods h f =
          if Int.equal nprods 0 then f None
          else Tacticals.tclZEROMSG (str"Not enough premisses") end
 
-let matches_pattern concl pat =
-  let matches env sigma =
-    match pat with
-    | None -> Proofview.tclUNIT ()
-    | Some pat ->
-       if Constr_matching.is_matching env sigma pat concl then
-         Proofview.tclUNIT ()
-       else
-         Tacticals.tclZEROMSG (str "pattern does not match")
-  in
-   Proofview.Goal.enter begin fun gl ->
-     let env = Proofview.Goal.env gl in
-     let sigma = Proofview.Goal.sigma gl in
-       matches env sigma end
-
 (** Semantics of type class resolution lemma application:
 
   - Use unification to find a well-typed substitution. There might
@@ -306,61 +253,39 @@ let rec e_trivial_fail_db only_classes db_list local_db secvars =
   tclSOLVE tacl
 
 and e_my_find_search db_list local_db secvars hdc complete only_classes env sigma concl0 =
-  let open Proofview.Notations in
-  let prods, concl = EConstr.decompose_prod_assum sigma concl0 in
+  let prods, concl = EConstr.decompose_prod_decls sigma concl0 in
   let nprods = List.length prods in
   let allowed_evars =
-    try
-      match hdc with
-      | Some (hd,_) when only_classes ->
-         let cl = Typeclasses.class_info env sigma hd in
-         if cl.cl_strict then
+    let all = Evarsolve.AllowedEvars.all in
+    match hdc with
+    | Some (hd,_) when only_classes ->
+      begin match Typeclasses.class_info hd with
+      | Some cl ->
+        if cl.cl_strict then
           let undefined = lazy (Evarutil.undefined_evars_of_term sigma concl) in
           let allowed evk = not (Evar.Set.mem evk (Lazy.force undefined)) in
           Evarsolve.AllowedEvars.from_pred allowed
-         else Evarsolve.AllowedEvars.all
-      | _ -> Evarsolve.AllowedEvars.all
-    with e when CErrors.noncritical e -> Evarsolve.AllowedEvars.all
+        else all
+      | None -> all
+      end
+    | _ -> all
   in
   let tac_of_hint =
     fun (flags, h) ->
-      let p = FullHint.pattern h in
       let name = FullHint.name h in
       let tac = function
         | Res_pf h ->
-           if get_typeclasses_filtered_unification () then
-             let tac =
-               with_prods nprods h
-                          (fun diff ->
-                             matches_pattern concl p <*>
-                               unify_resolve_refine flags h diff)
-             in Tacticals.tclTHEN tac Proofview.shelve_unifiable
-           else
-             let tac =
-               with_prods nprods h (unify_resolve ~with_evars:false flags h) in
-               Proofview.tclBIND (Proofview.with_shelf tac)
-                  (fun (gls, ()) -> shelve_dependencies gls)
+          let tac =
+            with_prods nprods h (unify_resolve ~with_evars:false flags h) in
+            Proofview.tclBIND (Proofview.with_shelf tac)
+              (fun (gls, ()) -> shelve_dependencies gls)
         | ERes_pf h ->
-           if get_typeclasses_filtered_unification () then
-             let tac = (with_prods nprods h
-                  (fun diff ->
-                             matches_pattern concl p <*>
-                             unify_resolve_refine flags h diff)) in
-             Tacticals.tclTHEN tac Proofview.shelve_unifiable
-           else
-             let tac =
-               with_prods nprods h (unify_resolve ~with_evars:true flags h) in
-               Proofview.tclBIND (Proofview.with_shelf tac)
-                  (fun (gls, ()) -> shelve_dependencies gls)
+          let tac =
+            with_prods nprods h (unify_resolve ~with_evars:true flags h) in
+            Proofview.tclBIND (Proofview.with_shelf tac)
+              (fun (gls, ()) -> shelve_dependencies gls)
         | Give_exact h ->
-           if get_typeclasses_filtered_unification () then
-             let tac =
-               matches_pattern concl p <*>
-                 Proofview.Goal.enter
-                   (fun gl -> unify_resolve_refine flags h None) in
-             Tacticals.tclTHEN tac Proofview.shelve_unifiable
-           else
-             e_give_exact flags h
+          e_give_exact flags h
       | Res_pf_THEN_trivial_fail h ->
          let fst = with_prods nprods h (unify_resolve ~with_evars:true flags h) in
          let snd = if complete then Tacticals.tclIDTAC
@@ -372,17 +297,11 @@ and e_my_find_search db_list local_db secvars hdc complete only_classes env sigm
       in
       let tac = FullHint.run h tac in
       let tac = if complete then Tacticals.tclCOMPLETE tac else tac in
-      let pp() =
-        match p with
-        | Some pat when get_typeclasses_filtered_unification () ->
-           str " with pattern " ++ Printer.pr_constr_pattern_env env sigma pat
-        | _ -> mt ()
-      in
       let extern = match FullHint.repr h with
         | Extern _ -> true
         | _ -> false
       in
-      (tac, FullHint.priority h, extern, name, lazy (FullHint.print env sigma h ++ pp ()))
+      (tac, FullHint.priority h, extern, name, lazy (FullHint.print env sigma h))
   in
   let hint_of_db = hintmap_of env sigma hdc secvars concl in
   let hintl = List.map_filter (fun db -> match hint_of_db db with
@@ -464,17 +383,18 @@ let is_unique env sigma concl =
 let top_sort evm undefs =
   let l' = ref [] in
   let tosee = ref undefs in
+  let cache = Evarutil.create_undefined_evars_cache () in
   let rec visit ev evi =
-    let evs = Evarutil.undefined_evars_of_evar_info evm evi in
+    let evs = Evarutil.filtered_undefined_evars_of_evar_info ~cache evm evi in
       tosee := Evar.Set.remove ev !tosee;
       Evar.Set.iter (fun ev ->
         if Evar.Set.mem ev !tosee then
-          visit ev (Evd.find evm ev)) evs;
+          visit ev (Evd.find_undefined evm ev)) evs;
       l' := ev :: !l';
   in
     while not (Evar.Set.is_empty !tosee) do
       let ev = Evar.Set.choose !tosee in
-        visit ev (Evd.find evm ev)
+        visit ev (Evd.find_undefined evm ev)
     done;
     List.rev !l'
 
@@ -492,16 +412,12 @@ let evars_to_goals p evm =
 let make_resolve_hyp env sigma st only_classes decl db =
   let id = NamedDecl.get_id decl in
   let cty = Evarutil.nf_evar sigma (NamedDecl.get_type decl) in
-  let rec iscl env ty =
-    let ctx, ar = decompose_prod_assum sigma ty in
+  let iscl env ty =
+    let ctx, ar = decompose_prod_decls sigma ty in
       match EConstr.kind sigma (fst (decompose_app sigma ar)) with
       | Const (c,_) -> is_class (GlobRef.ConstRef c)
       | Ind (i,_) -> is_class (GlobRef.IndRef i)
-      | _ ->
-          let env' = push_rel_context ctx env in
-          let ty' = Reductionops.whd_all env' sigma ar in
-               if not (EConstr.eq_constr sigma ty' ar) then iscl env' ty'
-               else false
+      | _ -> false
   in
   let is_class = iscl env cty in
   let keep = not only_classes || is_class in
@@ -526,6 +442,18 @@ let make_hints env sigma (modes,st) only_classes sign =
       else hints)
     sign db
 
+module Intpart = Unionfind.Make(Evar.Set)(Evar.Map)
+
+type solver = { solver :
+  Environ.env ->
+  Evd.evar_map ->
+  depth:int option ->
+  unique:bool ->
+  best_effort:bool ->
+  goals:Evar.t list ->
+  (bool * Evd.evar_map)
+}
+
 module Search = struct
   type autoinfo =
     { search_depth : int list;
@@ -551,7 +479,7 @@ module Search = struct
     let eq c1 c2 = EConstr.eq_constr sigma c1 c2 in
     if DirPath.equal cwd dir &&
          (onlyc == only_classes) &&
-           Context.Named.equal eq sign sign' &&
+           Context.Named.equal (ERelevance.equal sigma) eq sign sign' &&
              cached_modes == modes
     then cached_hints
     else
@@ -582,11 +510,18 @@ module Search = struct
     | _, _ -> e
 
   (** Determine if backtracking is needed for this goal.
-      If the type class is unique or in Prop
-      and there are no evars in the goal then we do
-      NOT backtrack. *)
-  let needs_backtrack env evd unique concl =
-    if unique || is_Prop env evd concl then
+      We generally backtrack except in the following (possibly
+      overlapping) cases:
+      - [unique_instances] is [true].
+        This is the case when the goal's class has [Unique Instances].
+      - [indep] is [true] and the current goal has no evars.
+        [indep] is generally [true] and only gets set to [false] if the
+        current goal's evar is mentioned in other goals.
+        ([indep] is the negation of [search_dep].)
+      - The current goal is a [Prop] and has no evars. *)
+  let needs_backtrack env evd ~unique_instances ~indep concl =
+    if unique_instances then false else
+    if indep || is_Prop env evd concl then
       occur_existential evd concl
     else true
 
@@ -608,7 +543,15 @@ module Search = struct
     else
       tclUNIT ()
 
-  let _ = CErrors.register_handler begin function
+  let pr_internal_exception ie =
+    match fst ie with
+    | ReachedLimit -> str "Proof-search reached its limit."
+    | NoApplicableHint -> str "Proof-search failed."
+    | StuckGoal | NonStuckFailure -> str "Proof-search got stuck."
+    | e -> CErrors.iprint ie
+
+  (* XXX Is this handler needed for something? *)
+  let () = CErrors.register_handler begin function
     | NonStuckFailure -> Some (str "NonStuckFailure")
     | NoApplicableHint -> Some (str "NoApplicableHint")
     | _ -> None
@@ -661,12 +604,12 @@ module Search = struct
             in
             cycle 1 (* Puts the first goal last *) <*>
             fixpoint progress tacs ((glid, ev, status, tac) :: stuck) fk (* Launches the search on the rest of the goals *)
-          | Fail (e, info) ->
+          | Fail ie ->
             let () = ppdebug 1 (fun () ->
                 str "Goal " ++ int glid ++ str" has no more solutions, returning exception: "
-                ++ CErrors.iprint (e, info))
+                ++ pr_internal_exception ie)
             in
-            fk (e, info)
+            fk ie
           | Next (res, fk') ->
             let () = ppdebug 1 (fun () ->
                 str "Goal " ++ int glid ++ str" has a success, continuing resolution")
@@ -755,8 +698,9 @@ module Search = struct
     let env = Goal.env gl in
     let concl = Goal.concl gl in
     let sigma = Goal.sigma gl in
-    let unique = not info.search_dep || is_unique env sigma concl in
-    let backtrack = needs_backtrack env sigma unique concl in
+    let unique_instances = is_unique env sigma concl in
+    let indep = not info.search_dep in
+    let backtrack = needs_backtrack env sigma ~unique_instances ~indep concl in
     let () = ppdebug 0 (fun () ->
         pr_depth info.search_depth ++ str": looking for " ++
         Printer.pr_econstr_env (Goal.env gl) sigma concl ++
@@ -789,18 +733,7 @@ module Search = struct
                  str" on" ++ spc () ++ pr_ev sigma (Proofview.Goal.goal gl)
                else mt ())
             in
-            let msg =
-              match fst ie with
-              | Pretype_errors.PretypeError (env, evd, Pretype_errors.CannotUnify (x,y,_)) ->
-                str"Cannot unify " ++
-                Printer.pr_econstr_env env evd x ++ str" and " ++
-                Printer.pr_econstr_env env evd y
-              | ReachedLimit -> str "Proof-search reached its limit."
-              | NoApplicableHint -> str "Proof-search failed."
-              | StuckGoal | NonStuckFailure -> str "Proof-search got stuck."
-              | e -> CErrors.iprint ie
-            in
-            (header ++ str " failed with: " ++ msg))
+            (header ++ str " failed with: " ++ pr_internal_exception ie))
       in
       let tac_of gls i j = Goal.enter begin fun gl' ->
         let sigma' = Goal.sigma gl' in
@@ -810,7 +743,7 @@ module Search = struct
         in
         let eq c1 c2 = EConstr.eq_constr sigma' c1 c2 in
         let hints' =
-          if b && not (Context.Named.equal eq (Goal.hyps gl') (Goal.hyps gl))
+          if b && not (Context.Named.equal (ERelevance.equal sigma) eq (Goal.hyps gl') (Goal.hyps gl))
           then
             let st = Hint_db.transparent_state info.search_hints in
             let modes = Hint_db.modes info.search_hints in
@@ -891,7 +824,7 @@ module Search = struct
         with_shelf res >>= fun (sh, ()) ->
         tclEVARMAP >>= finish sh
       in
-      if path_matches derivs [] then aux e tl
+      if path_matches_epsilon derivs then aux e tl
       else
         ortac
              (with_shelf tac >>= fun s ->
@@ -1061,19 +994,23 @@ module Search = struct
 
   let eauto_tac mst ?unique ~only_classes ~best_effort ?strategy ~depth ~dep hints =
     Hints.wrap_hint_warning @@
-      (eauto_tac_stuck mst ?unique ~only_classes
-          ~best_effort ?strategy ~depth ~dep hints)
+      eauto_tac_stuck mst ?unique ~only_classes
+          ~best_effort ?strategy ~depth ~dep hints
 
-  let run_on_goals env evm p tac goals nongoals =
-    let goalsl =
+  let preprocess_goals evm goals =
+    let sorted_goals =
       if get_typeclasses_dependency_order () then
         top_sort evm goals
       else Evar.Set.elements goals
     in
-    let goalsl = List.map Proofview.with_empty_state goalsl in
-    let tac = Proofview.Unsafe.tclNEWGOALS goalsl <*> tac in
     let evm = Evd.set_typeclass_evars evm Evar.Set.empty in
     let evm = Evd.push_future_goals evm in
+    evm, Evd.meta_list evm, sorted_goals
+
+
+  let run_on_goals env evm tac ~goals =
+    let goalsl = List.map Proofview.with_empty_state goals in
+    let tac = Proofview.Unsafe.tclNEWGOALS goalsl <*> tac in
     let _, pv = Proofview.init evm [] in
      (* Instance may try to call this before a proof is set up!
        Thus, give_me_the_proof will fail. Beware! *)
@@ -1098,62 +1035,52 @@ module Search = struct
     in
     let finished = Proofview.finished pv' in
     let evm' = Proofview.return pv' in
-    let shelf = Evd.shelf evm' in
-    assert(Evd.fold_undefined (fun ev _ acc ->
-        let okev = Evd.mem evm ev || List.mem ev shelf in
-          if not okev then
-            Feedback.msg_debug
-              (str "leaking evar " ++ int (Evar.repr ev) ++
-                  spc () ++ pr_ev_with_id evm' ev);
-          acc && okev) evm' true);
-    let _, evm' = Evd.pop_future_goals evm' in
+    (finished, evm')
+
+  let post_process_goals ~goals ~nongoals ~old_metas ~sigma ~finished =
+    let _, sigma = Evd.pop_future_goals sigma in
+    let tc_evars = Evd.get_typeclass_evars sigma in
     let () = ppdebug 1 (fun () ->
         str"Finished resolution with " ++ str(if finished then "a complete" else "an incomplete") ++
         str" solution." ++ fnl()  ++
         str"Old typeclass evars not concerned by this resolution = " ++
-        hov 0 (prlist_with_sep spc (pr_ev_with_id evm')
-                 (Evar.Set.elements (Evd.get_typeclass_evars evm'))) ++ fnl() ++
+        hov 0 (prlist_with_sep spc (pr_ev_with_id sigma)
+                 (Evar.Set.elements tc_evars)) ++ fnl() ++
         str"Shelf = " ++
-        hov 0 (prlist_with_sep spc (pr_ev_with_id evm')
-                 (Evar.Set.elements (Evd.get_typeclass_evars evm'))))
+        hov 0 (prlist_with_sep spc (pr_ev_with_id sigma)
+                 (Evar.Set.elements tc_evars)))
     in
-    let nongoals' =
-      Evar.Set.fold (fun ev acc -> match Evarutil.advance evm' ev with
-          | Some ev' -> Evar.Set.add ev acc
-          | None -> acc) (Evar.Set.union goals nongoals) (Evd.get_typeclass_evars evm')
+    let nongoals =
+      Evar.Set.fold (fun ev acc -> match Evarutil.advance sigma ev with
+          | Some ev -> Evar.Set.add ev acc
+          | None -> acc) (Evar.Set.union goals nongoals) tc_evars
     in
     (* FIXME: the need to merge metas seems to come from this being called
        internally from Unification. It should be handled there instead. *)
-    let evm' = Evd.meta_merge (Evd.meta_list evm) (Evd.clear_metas evm') in
-    let evm' = Evd.set_typeclass_evars evm' nongoals' in
+    let sigma = Evd.meta_merge old_metas (Evd.clear_metas sigma) in
+    let sigma = Evd.set_typeclass_evars sigma nongoals in
     let () = ppdebug 1 (fun () ->
         str"New typeclass evars are: " ++
-        hov 0 (prlist_with_sep spc (pr_ev_with_id evm') (Evar.Set.elements nongoals')))
+        hov 0 (prlist_with_sep spc (pr_ev_with_id sigma) (Evar.Set.elements nongoals)))
     in
-    Some (finished, evm')
+    sigma
 
-  let run_on_evars env evm p tac =
-    match evars_to_goals p evm with
-    | None -> None (* This happens only because there's no evar having p *)
-    | Some (goals, nongoals) ->
-      run_on_goals env evm p tac goals nongoals
-  let evars_eauto env evd depth only_classes ~best_effort unique dep mst hints p =
-    let eauto_tac = eauto_tac_stuck mst ~unique ~only_classes
-      ~best_effort
-      ~depth ~dep:(unique || dep) hints in
-    run_on_evars env evd p eauto_tac
 
-  let typeclasses_eauto env evd ?depth unique ~best_effort st hints p =
-    evars_eauto env evd depth true ~best_effort unique false st hints p
   (** Typeclasses eauto is an eauto which tries to resolve only
       goals of typeclass type, and assumes that the initially selected
       evars in evd are independent of the rest of the evars *)
+  let typeclasses_eauto env evd ~goals ?depth ~unique ~best_effort st hints =
+    let only_classes = true in
+    let dep = unique in
+    NewProfile.profile "typeclass search" (fun () ->
+      run_on_goals env evd (eauto_tac_stuck st ~unique ~only_classes ~best_effort ~depth ~dep hints) ~goals) ()
 
-  let typeclasses_resolve env evd depth unique ~best_effort p =
+  let typeclasses_resolve : solver = { solver = fun env evd ~depth ~unique ~best_effort ~goals ->
     let db = searchtable_map typeclasses_db in
     let st = Hint_db.transparent_state db in
     let modes = Hint_db.modes db in
-    typeclasses_eauto env evd ?depth ~best_effort unique (modes,st) [db] p
+    typeclasses_eauto env evd ~goals ?depth ~best_effort ~unique (modes,st) [db]
+  }
 end
 
 let typeclasses_eauto ?(only_classes=false)
@@ -1180,8 +1107,6 @@ let typeclasses_eauto ?(only_classes=false)
     Beware of the imperative effects on the partition structure,
     it should not be shared, but only used locally. *)
 
-module Intpart = Unionfind.Make(Evar.Set)(Evar.Map)
-
 let deps_of_constraints cstrs evm p =
   List.iter (fun (_, _, x, y) ->
     let evx = Evarutil.undefined_evars_of_term evm x in
@@ -1211,7 +1136,7 @@ let is_inference_forced p evd ev =
   try
     if Evar.Set.mem ev (Evd.get_typeclass_evars evd) && p ev
     then
-      let (loc, k) = evar_source (Evd.find evd ev) in
+      let (loc, k) = evar_source (Evd.find_undefined evd ev) in
       match k with
         | Evar_kinds.ImplicitArg (_, _, b) -> b
         | Evar_kinds.QuestionMark _ -> false
@@ -1229,7 +1154,7 @@ let is_mandatory p comp evd =
 let select_and_update_evars p oevd in_comp evd ev =
   try
     if Evd.is_typeclass_evar oevd ev then
-      (in_comp ev && p evd ev (Evd.find evd ev))
+      (in_comp ev && p evd ev (Evd.find_undefined evd ev))
     else false
   with Not_found -> false
 
@@ -1244,23 +1169,53 @@ let find_undefined p oevd evd =
 
 exception Unresolved of evar_map
 
+type condition = Environ.env -> evar_map -> Evar.Set.t -> bool
+
+type tc_solver = solver * condition
+
+let class_solvers = ref (CString.Map.empty : tc_solver CString.Map.t)
+
+let register_solver ~name ?(override=false) h =
+  if not override && CString.Map.mem name !class_solvers then
+    CErrors.anomaly ~label:"Class_tactics.register_solver"
+      Pp.(str (Printf.sprintf {|Solver "%s" is already registered|} name));
+  class_solvers := CString.Map.add name h !class_solvers
+
+let active_solvers = Summary.ref ~name:"typeclass_solvers" ([] : string list)
+
+let deactivate_solver ~name =
+  active_solvers := List.filter (fun s -> not (String.equal s name)) !active_solvers
+
+let activate_solver ~name =
+  assert (CString.Map.mem name !class_solvers);
+  deactivate_solver ~name;
+  active_solvers := name :: !active_solvers
+
+let find_solver env evd (s : Intpart.set) =
+  let rec find_solver = function
+    | [] -> Search.typeclasses_resolve
+    | hd :: tl ->
+      try
+        let (solver,cond) = CString.Map.find hd !class_solvers in
+        if cond env evd s then solver else find_solver tl
+      with Not_found ->  find_solver tl in
+  find_solver !active_solvers
+
 (** If [do_split] is [true], we try to separate the problem in
     several components and then solve them separately *)
-let resolve_all_evars depth unique env p oevd do_split fail =
+let resolve_all_evars depth unique env p oevd fail =
   let () =
     ppdebug 0 (fun () ->
         str"Calling typeclass resolution with flags: "++
         str"depth = " ++ (match depth with None -> str "∞" | Some d -> int d) ++ str"," ++
         str"unique = " ++ bool unique ++ str"," ++
-        str"do_split = " ++ bool do_split ++ str"," ++
         str"fail = " ++ bool fail);
     ppdebug 2 (fun () ->
         str"Initial evar map: " ++
         Termops.pr_evar_map ~with_univs:!Detyping.print_universes None env oevd)
   in
-  let tcs = Evd.get_typeclass_evars oevd in
-  let split = if do_split then split_evars p oevd else [tcs] in
-  let in_comp comp ev = if do_split then Evar.Set.mem ev comp else true in
+  let split = split_evars p oevd in
+  let in_comp comp ev = Evar.Set.mem ev comp in
   let rec docomp evd = function
     | [] ->
       let () = ppdebug 2 (fun () ->
@@ -1272,26 +1227,28 @@ let resolve_all_evars depth unique env p oevd do_split fail =
       let p = select_and_update_evars p oevd (in_comp comp) in
       try
         (try
-          let res = Search.typeclasses_resolve env evd depth
-            ~best_effort:true unique p in
-          match res with
-          | Some (finished, evd') ->
-            if has_undefined p oevd evd' then
+          (* evars_to_goals p evd gives none when there's no evar having p *)
+          match evars_to_goals p evd with
+          | Some (goals, nongoals) ->
+            let solver = find_solver env evd comp in
+            let evd, old_metas, sorted_goals = Search.preprocess_goals evd goals in
+            let finished, evd = solver.solver env evd ~goals:sorted_goals ~best_effort:true ~depth ~unique in
+            let evd = Search.post_process_goals ~goals ~nongoals ~old_metas ~sigma:evd ~finished in
+            if has_undefined p oevd evd then
               let () = if finished then ppdebug 1 (fun () ->
                   str"Proof is finished but there remain undefined evars: " ++
-                  prlist_with_sep spc (pr_ev evd')
-                    (Evar.Set.elements (find_undefined p oevd evd')))
+                  prlist_with_sep spc (pr_ev evd)
+                    (Evar.Set.elements (find_undefined p oevd evd)))
               in
-              raise (Unresolved evd')
-            else docomp evd' comps
+              raise (Unresolved evd)
+            else docomp evd comps
           | None -> docomp evd comps (* No typeclass evars left in this component *)
         with Not_found ->
           (* Typeclass resolution failed *)
           raise (Unresolved evd))
       with Unresolved evd' ->
-        if fail && (not do_split || is_mandatory (p evd') comp evd')
+        if fail && is_mandatory (p evd') comp evd'
         then (* Unable to satisfy the constraints. *)
-          let comp = if do_split then Some comp else None in
           error_unresolvable env evd' comp
         else (* Best effort: use the best found solution on this component *)
           docomp evd' comps
@@ -1309,20 +1266,20 @@ let classes_transparent_state () =
   try Hint_db.transparent_state (searchtable_map typeclasses_db)
   with Not_found -> TransparentState.empty
 
-let resolve_typeclass_evars depth unique env evd filter split fail =
+let resolve_typeclass_evars depth unique env evd filter fail =
   let evd =
     try Evarconv.solve_unif_constraints_with_heuristics
       ~flags:(Evarconv.default_flags_of (classes_transparent_state())) env evd
     with e when CErrors.noncritical e -> evd
   in
     resolve_all_evars depth unique env
-                      (initial_select_evars filter) evd split fail
+      (initial_select_evars filter) evd fail
 
-let solve_inst env evd filter unique split fail =
+let solve_inst env evd filter unique fail =
   let ((), sigma) = Hints.wrap_hint_warning_fun env evd begin fun evd ->
     (), resolve_typeclass_evars
     (get_typeclasses_depth ())
-    unique env evd filter split fail
+    unique env evd filter fail
   end in
   sigma
 
@@ -1397,9 +1354,21 @@ let autoapply c i =
   let cty = Tacmach.pf_get_type_of gl c in
   let env = Proofview.Goal.env gl in
   let sigma = Proofview.Goal.sigma gl in
-  let ce = mk_clenv_from env sigma (c,cty) in
+  let ce = Clenv.mk_clenv_from env sigma (c,cty) in
   Clenv.res_pf ~with_evars:true ~with_classes:false ~flags ce <*>
       Proofview.tclEVARMAP >>= (fun sigma ->
       let sigma = Typeclasses.make_unresolvables
-          (fun ev -> Typeclasses.all_goals ev (Lazy.from_val (snd (Evd.evar_source (Evd.find sigma ev))))) sigma in
+          (fun ev -> Typeclasses.all_goals ev (Lazy.from_val (snd (Evd.evar_source (Evd.find_undefined sigma ev))))) sigma in
       Proofview.Unsafe.tclEVARS sigma) end
+
+let resolve_tc c =
+  let open Proofview.Notations in
+  Proofview.tclENV >>= fun env ->
+  Proofview.tclEVARMAP >>= fun sigma ->
+  let depth = get_typeclasses_depth () in
+  let unique = get_typeclasses_unique_solutions () in
+  let evars = Evarutil.undefined_evars_of_term sigma c in
+  let filter = (fun ev _ -> Evar.Set.mem ev evars) in
+  let fail = true in
+  let sigma = resolve_all_evars depth unique env (initial_select_evars filter) sigma fail in
+  Proofview.Unsafe.tclEVARS sigma
