@@ -33,19 +33,35 @@ type individual_scheme_object_function =
 
 (** Main functions to register a scheme builder. Note these functions
    are not safe to be used by plugins as their effects won't be undone
-   on backtracking *)
+   on backtracking.
+
+    In [declare_X_scheme_object key ?suff ?deps f], [key] is the name
+    of the scheme kind. It must be unique across the Coq process's
+    lifetime. It is used to generate [scheme_kind] in a marshal-stable
+    way and as the scheme name in Register Scheme.
+
+    [suff] defaults to [key], generated schemes which aren't given an
+    explicit name will be named "ind_suff" where "ind" is the
+    inductive's name.
+*)
 
 val declare_mutual_scheme_object : string ->
+  ?suff:string ->
   ?deps:(Environ.env -> MutInd.t -> scheme_dependency list) ->
-  ?aux:string ->
   mutual_scheme_object_function -> mutual scheme_kind
 
 val declare_individual_scheme_object : string ->
+  ?suff:string ->
   ?deps:(Environ.env -> inductive -> scheme_dependency list) ->
-  ?aux:string ->
   individual_scheme_object_function ->
   individual scheme_kind
 
+val is_declared_scheme_object : string -> bool
+(** Is the string used as the name of a [scheme_kind]? *)
+
+val scheme_kind_name : _ scheme_kind -> string
+(** Name of a [scheme_kind]. Can be used to register with DeclareScheme. *)
+
 (** Force generation of a (mutually) scheme with possibly user-level names *)
 
 val define_individual_scheme : ?loc:Loc.t -> individual scheme_kind ->
