@@ -91,13 +91,17 @@ type compilation_unit_name = DirPath.t
 
 type library_disk = {
   md_compiled : Safe_typing.compiled_library;
+  md_syntax_objects : library_objects;
   md_objects : library_objects;
 }
 
+type library_info
+
 type summary_disk = {
   md_name : compilation_unit_name;
   md_deps : (compilation_unit_name * Safe_typing.vodigest) array;
   md_ocaml : string;
+  md_info : library_info;
 }
 
 type library_t = {
@@ -106,7 +110,7 @@ type library_t = {
   library_data : Safe_typing.compiled_library;
   library_deps : (compilation_unit_name * Safe_typing.vodigest) array;
   library_digests : Safe_typing.vodigest;
-  library_extra_univs : Univ.ContextSet.t;
+  library_vm : Vmlibrary.on_disk;
 }
 
 let libraries_table : string DPmap.t ref = ref DPmap.empty
@@ -118,30 +122,28 @@ let register_loaded_library senv libname file =
   let () = Nativecode.register_native_file prefix in
   senv
 
-let mk_library sd f md digests univs =
+let mk_library sd f md digests vm =
   {
     library_name     = sd.md_name;
     library_file     = f;
     library_data     = md;
     library_deps     = sd.md_deps;
     library_digests  = digests;
-    library_extra_univs = univs;
+    library_vm = vm;
   }
 
+let summary_seg : summary_disk ObjFile.id = ObjFile.make_id "summary"
+let library_seg : library_disk ObjFile.id = ObjFile.make_id "library"
+
 let intern_from_file f =
   let ch = System.with_magic_number_check (fun file -> ObjFile.open_in ~file) f in
-  let (lsd : summary_disk), digest_lsd = ObjFile.marshal_in_segment ch ~segment:"summary" in
-  let ((lmd : library_disk), digest_lmd) = ObjFile.marshal_in_segment ch ~segment:"library" in
-  let (univs : (Univ.ContextSet.t * bool) option), digest_u = ObjFile.marshal_in_segment ch ~segment:"universes" in
+  let lsd, digest_lsd = ObjFile.marshal_in_segment ch ~segment:summary_seg in
+  let lmd, digest_lmd = ObjFile.marshal_in_segment ch ~segment:library_seg in
+  let vmlib = Vmlibrary.load ~file:f lsd.md_name ch in
   ObjFile.close_in ch;
   System.check_caml_version ~caml:lsd.md_ocaml ~file:f;
   let open Safe_typing in
-  match univs with
-  | None -> mk_library lsd f lmd.md_compiled (Dvo_or_vi digest_lmd) Univ.ContextSet.empty
-  | Some (uall,true) ->
-      mk_library lsd f lmd.md_compiled (Dvivo (digest_lmd,digest_u)) uall
-  | Some (_,false) ->
-      mk_library lsd f lmd.md_compiled (Dvo_or_vi digest_lmd) Univ.ContextSet.empty
+  mk_library lsd f lmd.md_compiled (Dvo_or_vi digest_lmd) vmlib
 
 let rec intern_library (needed, contents) dir =
   (* Look if already listed and consequently its dependencies too *)
@@ -171,7 +173,7 @@ and intern_mandatory_library caller from libs (dir,d) =
 
 let register_library senv m =
   let mp = MPfile m.library_name in
-  let mp', senv = Safe_typing.import m.library_data m.library_extra_univs m.library_digests senv in
+  let mp', senv = Safe_typing.import m.library_data m.library_vm m.library_digests senv in
   let () =
     if not (ModPath.equal mp mp') then
       anomaly (Pp.str "Unexpected disk module name.")
@@ -207,7 +209,7 @@ let convert_string d =
   with CErrors.UserError _ ->
     Flags.if_verbose Feedback.msg_warning
       (str "Directory " ++ str d ++ str " cannot be used as a Coq identifier (skipped)");
-    raise Exit
+    raise_notrace Exit
 
 let coq_root = Id.of_string "Coq"
 
@@ -264,7 +266,6 @@ let fb_handler = function
 
 let init_coq () =
   let senv = Safe_typing.empty_environment in
-  let () = Flags.set_native_compiler true in
   let senv = Safe_typing.set_native_compiler true senv in
   let () = Safe_typing.allow_delayed_constants := false in
   let dummy = Names.DirPath.make [Names.Id.of_string_soft "@native"] in
@@ -347,7 +348,7 @@ let rec parse_args (args : string list) accu =
     Nativelib.output_dir := dir;
     parse_args rem accu
   | "-coqlib" :: s :: rem ->
-    if not (Minisys.exists_dir s) then
+    if not (System.exists_dir s) then
       fatal_error (str "Directory '" ++ str s ++ str "' does not exist") false;
     Boot.Env.set_coqlib s;
     parse_args rem accu
