@@ -11,7 +11,7 @@
 open Names
 open CErrors
 
-let extra_deps = Summary.ref ~name:"extra_deps" Id.Map.empty
+let extra_deps = Summary.ref ~name:"extra_deps" ~stage:Summary.Stage.Synterp Id.Map.empty
 
 let bind_extra_dep ?loc path id =
   match Id.Map.find_opt id !extra_deps with
