@@ -15,39 +15,25 @@ open Vernacexpr
 
 (** Entry points for the vernacular commands Fixpoint and CoFixpoint *)
 
-val do_fixpoint_interactive
-  : scope:Locality.definition_scope
+val do_fixpoint
+  : ?scope:Locality.definition_scope
   -> ?clearbody:bool
   -> poly:bool
   -> ?typing_flags:Declarations.typing_flags
-  -> ?deprecation:Deprecation.t
-  -> fixpoint_expr list
-  -> Declare.Proof.t
-
-val do_fixpoint
-   : ?scope:Locality.definition_scope
-   -> ?clearbody:bool
-  -> poly:bool
-  -> ?typing_flags:Declarations.typing_flags
-  -> ?deprecation:Deprecation.t
+  -> ?user_warns:UserWarn.t
   -> ?using:Vernacexpr.section_subset_expr
   -> fixpoint_expr list
-  -> unit
-
-val do_cofixpoint_interactive
-  : scope:Locality.definition_scope
-  -> poly:bool
-  -> ?deprecation:Deprecation.t
-  -> cofixpoint_expr list
-  -> Declare.Proof.t
+  -> Declare.Proof.t option
 
 val do_cofixpoint
-  : scope:Locality.definition_scope
+  : ?scope:Locality.definition_scope
+  -> ?clearbody:bool
   -> poly:bool
-  -> ?deprecation:Deprecation.t
+  -> ?typing_flags:Declarations.typing_flags
+  -> ?user_warns:UserWarn.t
   -> ?using:Vernacexpr.section_subset_expr
   -> cofixpoint_expr list
-  -> unit
+  -> Declare.Proof.t option
 
 (************************************************************************)
 (** Internal API  *)
@@ -62,10 +48,10 @@ val adjust_rec_order
   -> lident option
 
 (** names / relevance / defs / types *)
-type ('constr, 'types) recursive_preentry = Id.t list * Sorts.relevance list * 'constr option list * 'types list
+type ('constr, 'types, 'r) recursive_preentry = Id.t list * 'r list * 'constr option list * 'types list
 
 (** Exported for Program *)
-val interp_recursive :
+val interp_recursive_evars :
   Environ.env ->
   (* Misc arguments *)
   program_mode:bool -> cofix:bool ->
@@ -74,21 +60,21 @@ val interp_recursive :
   (* env / signature / univs / evar_map *)
   (Environ.env * EConstr.named_context * UState.universe_decl * Evd.evar_map) *
   (* names / defs / types *)
-  (EConstr.t, EConstr.types) recursive_preentry *
+  (EConstr.t, EConstr.types, EConstr.ERelevance.t) recursive_preentry *
   (* ctx per mutual def / implicits / struct annotations *)
   (EConstr.rel_context * Impargs.manual_implicits * int option) list
 
 (** Exported for Funind *)
 
-val interp_fixpoint
+val interp_recursive
   :  ?check_recursivity:bool
   -> ?typing_flags:Declarations.typing_flags
   -> cofix:bool
   -> lident option fix_expr_gen list
-  -> (Constr.t, Constr.types) recursive_preentry *
+  -> (Constr.t, Constr.types, Sorts.relevance) recursive_preentry *
      UState.universe_decl * UState.t *
      (EConstr.rel_context * Impargs.manual_implicits * int option) list
 
 (** Very private function, do not use *)
 val compute_possible_guardness_evidences :
-  ('a, 'b) Context.Rel.pt * 'c * int option -> int list
+  ('a, 'b, 'r) Context.Rel.pt * 'c * int option -> int list
