@@ -14,6 +14,7 @@ open Names
 open Libnames
 open Constrexpr
 open Extend
+open Notationextern
 open Notation_gram
 open Pcoq
 
@@ -33,7 +34,7 @@ open Pcoq
 let constr_level = string_of_int
 
 let default_levels =
-  [200,Gramlib.Gramext.RightA,false;
+  [200,Gramlib.Gramext.RightA,true;
    100,Gramlib.Gramext.RightA,false;
    99,Gramlib.Gramext.RightA,true;
    90,Gramlib.Gramext.RightA,true;
@@ -205,21 +206,22 @@ let assoc_eq al ar =
   | LeftA, LeftA -> true
   | _, _ -> false
 
-(** [adjust_level assoc from prod] where [assoc] and [from] are the name
+(** [adjust_level assoc fromlev prod] where [assoc] and [fromlev] are the name
    and associativity of the level where to add the rule; the meaning of
    the result is
 
      DefaultLevel = entry name
      NextLevel = NEXT
      NumLevel n = constr LEVEL n *)
-let adjust_level custom assoc (custom',from) p = let open Gramlib.Gramext in match p with
+let adjust_level custom assoc {notation_entry = custom'; notation_level = fromlev} p =
+  let open Gramlib.Gramext in match p with
 (* If a level in a different grammar, no other choice than denoting it by absolute level *)
-  | (NumLevel n,_) when not (Notation.notation_entry_eq custom custom') -> NumLevel n
+  | (NumLevel n,_) when not (notation_entry_eq custom custom') -> NumLevel n
 (* If a default level in a different grammar, the entry name is ok *)
   | (DefaultLevel,InternalProd) ->
-    if Notation.notation_entry_eq custom InConstrEntry then NumLevel 200 else DefaultLevel
-  | (DefaultLevel,BorderProd _) when not (Notation.notation_entry_eq custom custom') ->
-    if Notation.notation_entry_eq custom InConstrEntry then NumLevel 200 else DefaultLevel
+    if notation_entry_eq custom InConstrEntry then NumLevel 200 else DefaultLevel
+  | (DefaultLevel,BorderProd _) when not (notation_entry_eq custom custom') ->
+    if notation_entry_eq custom InConstrEntry then NumLevel 200 else DefaultLevel
 (* Associativity is None means force the level *)
   | (NumLevel n,BorderProd (_,None)) -> NumLevel n
   | (DefaultLevel,BorderProd (_,None)) -> assert false
@@ -228,7 +230,7 @@ let adjust_level custom assoc (custom',from) p = let open Gramlib.Gramext in mat
   | ((NumLevel _ | DefaultLevel),BorderProd (Right,Some (NonA|LeftA))) -> NextLevel
   (* If RightA on the right-hand side, set to the explicit (current) level *)
   | (NumLevel n,BorderProd (Right,Some RightA)) -> NumLevel n
-  | (DefaultLevel,BorderProd (Right,Some RightA)) -> NumLevel from
+  | (DefaultLevel,BorderProd (Right,Some RightA)) -> NumLevel fromlev
 (* Compute production name on the left side *)
   (* If NonA on the left-hand side, adopt the current assoc ?? *)
   | ((NumLevel _ | DefaultLevel),BorderProd (Left,Some NonA)) -> DefaultLevel
@@ -239,10 +241,10 @@ let adjust_level custom assoc (custom',from) p = let open Gramlib.Gramext in mat
   | (NumLevel n,BorderProd (Left,Some LeftA)) -> NumLevel n
   | ((NumLevel _ | DefaultLevel),BorderProd (Left,Some _)) -> NextLevel
   (* None means NEXT *)
-  | (NextLevel,_) -> assert (Notation.notation_entry_eq custom custom'); NextLevel
+  | (NextLevel,_) -> assert (notation_entry_eq custom custom'); NextLevel
 (* Compute production name elsewhere *)
   | (NumLevel n,InternalProd) ->
-    if from = n + 1 then NextLevel else NumLevel n
+    if fromlev = n + 1 then NextLevel else NumLevel n
 
 type _ target =
 | ForConstr : constr_expr target
@@ -253,21 +255,22 @@ type prod_info = production_level * production_position
 type (_, _) entry =
 | TTIdent : ('self, lident) entry
 | TTName : ('self, lname) entry
-| TTGlobal : ('self, qualid) entry
-| TTBigint : ('self, string) entry
+| TTGlobal : ('r, qualid) entry
+| TTBigint : ('r, string) entry
 | TTBinder : bool -> ('self, kinded_cases_pattern_expr) entry
 | TTConstr : notation_entry * prod_info * 'r target -> ('r, 'r) entry
 | TTConstrList : notation_entry * prod_info * (bool * string) list * 'r target -> ('r, 'r list) entry
 | TTPattern : int -> ('self, cases_pattern_expr) entry
 | TTOpenBinderList : ('self, local_binder_expr list) entry
-| TTClosedBinderList : (bool * string) list -> ('self, local_binder_expr list list) entry
+| TTClosedBinderListPure : (bool * string) list -> ('self, local_binder_expr list list) entry
+| TTClosedBinderListOther : ('self, 'a) entry * (bool * string) list -> ('self, 'a list) entry
 
 type _ any_entry = TTAny : ('s, 'r) entry -> 's any_entry
 
 let constr_custom_entry : (string, Constrexpr.constr_expr) entry_command =
-  create_entry_command "constr" (fun s st -> [s], st)
+  create_entry_command "constr" { eext_fun = (fun s st -> [s], st); eext_eq = (==) (* FIXME *) }
 let pattern_custom_entry : (string, Constrexpr.cases_pattern_expr) entry_command =
-  create_entry_command "pattern" (fun s st -> [s], st)
+  create_entry_command "pattern" { eext_fun = (fun s st -> [s], st); eext_eq = (==) (* FIXME *) }
 
 let custom_entry_locality = Summary.ref ~name:"LOCAL-CUSTOM-ENTRY" String.Set.empty
 (** If the entry is present then local *)
@@ -290,13 +293,13 @@ let find_custom_entry s =
 
 let exists_custom_entry s = match find_custom_entry s with
 | _ -> true
-| exception _ -> false
+| exception e when CErrors.noncritical e -> false
 
 let locality_of_custom_entry s = String.Set.mem s !custom_entry_locality
 
-(* This computes the name of the level where to add a new rule *)
-let interp_constr_entry_key : type r. _ -> r target -> int -> r Entry.t * int option =
-  fun custom forpat level ->
+(** This computes the name of the level where to add a new rule *)
+let interp_constr_entry_key : type r. _ -> r target -> r Entry.t * int option =
+  fun {notation_entry = custom; notation_level = level} forpat ->
   match custom with
   | InCustomEntry s ->
      (let (entry_for_constr, entry_for_patttern) = find_custom_entry s in
@@ -321,20 +324,25 @@ let target_entry : type s. notation_entry -> s target -> s Entry.t = function
    | ForConstr -> entry_for_constr
    | ForPattern -> entry_for_patttern
 
-let is_self custom (custom',from) e = Notation.notation_entry_eq custom custom' && match e with
-| (NumLevel n, BorderProd (Right, _ (* Some(NonA|LeftA) *))) -> false
-| (NumLevel n, BorderProd (Left, _)) -> Int.equal from n
-| _ -> false
+let is_self custom {notation_entry = custom'; notation_level = fromlev} e =
+  notation_entry_eq custom custom' && match e with
+  | (NumLevel n, BorderProd (Right, _ (* Some(NonA|LeftA) *))) -> false
+  | (NumLevel n, BorderProd (Left, _)) -> Int.equal fromlev n
+  | _ -> false
 
-let is_binder_level custom (custom',from) e = match e with
-| (NumLevel 200, (BorderProd (Right, _) | InternalProd)) ->
-  custom = InConstrEntry && custom' = InConstrEntry && from = 200
-| _ -> false
+let is_binder_level custom {notation_entry = custom'; notation_level = fromlev} e =
+  match e with
+  | (NumLevel 200, (BorderProd (Right, _) | InternalProd)) ->
+    custom = InConstrEntry && custom' = InConstrEntry && fromlev = 200
+  | _ -> false
 
 let make_pattern (keyword,s) =
    if keyword then TPattern (Tok.PKEYWORD s) else
      match NumTok.Unsigned.parse_string s with
      | Some n -> TPattern (Tok.PNUMBER (Some n))
+     | None ->
+     match String.unquote_coq_string s with
+     | Some s -> TPattern (Tok.PSTRING (Some s))
      | None -> TPattern (Tok.PIDENT (Some s))
 
 let make_sep_rules tkl =
@@ -359,7 +367,7 @@ let symbol_of_target : type s. _ -> _ -> _ -> _ -> s target -> (s, s) mayrec_sym
     | NumLevel lev -> MayRecNo (Pcoq.Symbol.nterml g (string_of_int lev))
     end
 
-let symbol_of_entry : type s r. _ -> _ -> (s, r) entry -> (s, r) mayrec_symbol = fun assoc from typ -> match typ with
+let rec symbol_of_entry : type s r. _ -> _ -> (s, r) entry -> (s, r) mayrec_symbol = fun assoc from typ -> match typ with
 | TTConstr (s, p, forpat) -> symbol_of_target s p assoc from forpat
 | TTConstrList (s, typ', [], forpat) ->
   begin match symbol_of_target s typ' assoc from forpat with
@@ -370,17 +378,25 @@ let symbol_of_entry : type s r. _ -> _ -> (s, r) entry -> (s, r) mayrec_symbol =
   | MayRecNo s -> MayRecNo (Pcoq.Symbol.list1sep s (make_sep_rules tkl) false)
   | MayRecMay s -> MayRecMay (Pcoq.Symbol.list1sep s (make_sep_rules tkl) false) end
 | TTPattern p -> MayRecNo (Pcoq.Symbol.nterml Constr.pattern (string_of_int p))
-| TTClosedBinderList [] -> MayRecNo (Pcoq.Symbol.list1 (Pcoq.Symbol.nterm Constr.binder))
-| TTClosedBinderList tkl -> MayRecNo (Pcoq.Symbol.list1sep (Pcoq.Symbol.nterm Constr.binder) (make_sep_rules tkl) false)
+| TTOpenBinderList -> MayRecNo (Pcoq.Symbol.nterm Constr.open_binders)
+| TTClosedBinderListPure [] -> MayRecNo (Pcoq.Symbol.list1 (Pcoq.Symbol.nterm Constr.binder))
+| TTClosedBinderListPure tkl -> MayRecNo (Pcoq.Symbol.list1sep (Pcoq.Symbol.nterm Constr.binder) (make_sep_rules tkl) false)
+| TTClosedBinderListOther (typ,[]) ->
+  begin match symbol_of_entry assoc from typ with
+  | MayRecNo s -> MayRecNo (Pcoq.Symbol.list1 s)
+  | MayRecMay s -> MayRecMay (Pcoq.Symbol.list1 s) end
+| TTClosedBinderListOther (typ,tkl) ->
+  begin match symbol_of_entry assoc from typ with
+  | MayRecNo s -> MayRecNo (Pcoq.Symbol.list1sep s (make_sep_rules tkl) false)
+  | MayRecMay s -> MayRecMay (Pcoq.Symbol.list1sep s (make_sep_rules tkl) false) end
 | TTIdent -> MayRecNo (Pcoq.Symbol.nterm Prim.identref)
 | TTName -> MayRecNo (Pcoq.Symbol.nterm Prim.name)
 | TTBinder true -> MayRecNo (Pcoq.Symbol.nterm Constr.one_open_binder)
 | TTBinder false -> MayRecNo (Pcoq.Symbol.nterm Constr.one_closed_binder)
-| TTOpenBinderList -> MayRecNo (Pcoq.Symbol.nterm Constr.open_binders)
 | TTBigint -> MayRecNo (Pcoq.Symbol.nterm Prim.bignat)
 | TTGlobal -> MayRecNo (Pcoq.Symbol.nterm Constr.global)
 
-let interp_entry forpat e = match e with
+let rec interp_entry forpat e = match e with
 | ETProdIdent -> TTAny TTIdent
 | ETProdName -> TTAny TTName
 | ETProdGlobal -> TTAny TTGlobal
@@ -390,7 +406,9 @@ let interp_entry forpat e = match e with
 | ETProdPattern p -> TTAny (TTPattern p)
 | ETProdConstrList (s, p, tkl) -> TTAny (TTConstrList (s, p, tkl, forpat))
 | ETProdBinderList ETBinderOpen -> TTAny TTOpenBinderList
-| ETProdBinderList (ETBinderClosed tkl) -> TTAny (TTClosedBinderList tkl)
+| ETProdBinderList (ETBinderClosed (None, tkl)) -> TTAny (TTClosedBinderListPure tkl)
+| ETProdBinderList (ETBinderClosed (Some e, tkl)) ->
+  let TTAny e = interp_entry forpat e in TTAny (TTClosedBinderListOther (e, tkl))
 
 let cases_pattern_expr_of_id { CAst.loc; v = id } =
   CAst.make ?loc @@ CPatAtom (Some (qualid_of_ident ?loc id))
@@ -411,24 +429,16 @@ let push_constr subst v = { subst with constrs = v :: subst.constrs }
 let push_item : type s r. s target -> (s, r) entry -> s env -> r -> s env = fun forpat e subst v ->
 match e with
 | TTConstr _ -> push_constr subst v
-| TTIdent ->
-  begin match forpat with
-  | ForConstr -> { subst with binders = (cases_pattern_expr_of_id v, Glob_term.Explicit) :: subst.binders }
-  | ForPattern -> push_constr subst (cases_pattern_expr_of_id v)
-  end
-| TTName ->
-  begin match forpat with
-  | ForConstr -> { subst with binders = (cases_pattern_expr_of_name v, Glob_term.Explicit) :: subst.binders }
-  | ForPattern -> push_constr subst (cases_pattern_expr_of_name v)
-  end
-| TTPattern _ ->
-  begin match forpat with
-  | ForConstr -> { subst with binders = (v, Glob_term.Explicit) :: subst.binders }
-  | ForPattern -> push_constr subst v
-  end
+| TTIdent -> { subst with binders = (cases_pattern_expr_of_id v, Glob_term.Explicit) :: subst.binders }
+| TTName -> { subst with binders = (cases_pattern_expr_of_name v, Glob_term.Explicit) :: subst.binders }
+| TTPattern _ -> { subst with binders = (v, Glob_term.Explicit) :: subst.binders }
 | TTBinder o -> { subst with binders = v :: subst.binders }
 | TTOpenBinderList -> { subst with binderlists = v :: subst.binderlists }
-| TTClosedBinderList _ -> { subst with binderlists = List.flatten v :: subst.binderlists }
+| TTClosedBinderListPure _ -> { subst with binderlists = List.flatten v :: subst.binderlists }
+| TTClosedBinderListOther (TTIdent, _) -> { subst with binderlists = List.map (fun a -> CLocalPattern (cases_pattern_expr_of_id a)) v :: subst.binderlists }
+| TTClosedBinderListOther (TTName, _) -> { subst with binderlists = List.map (fun a -> CLocalPattern (cases_pattern_expr_of_name a)) v :: subst.binderlists }
+| TTClosedBinderListOther (TTPattern _, _) -> { subst with binderlists = List.map (fun a -> CLocalPattern a) v :: subst.binderlists }
+| TTClosedBinderListOther _ -> user_err (Pp.str "Invalid binder list entry.")
 | TTBigint ->
   begin match forpat with
   | ForConstr ->  push_constr subst (CAst.make @@ CPrim (Number (NumTok.Signed.of_int_string v)))
@@ -542,7 +552,7 @@ let prepare_empty_levels forpat (where,(pos,p4assoc,name,reinit)) =
 let different_levels (custom,opt_level) (custom',string_level) =
   match opt_level with
   | None -> true
-  | Some level -> not (Notation.notation_entry_eq custom custom') || level <> int_of_string string_level
+  | Some level -> not (notation_entry_eq custom custom') || level <> int_of_string string_level
 
 let rec pure_sublevels' assoc from forpat level = function
 | [] -> []
@@ -554,7 +564,7 @@ let rec pure_sublevels' assoc from forpat level = function
        (match Pcoq.level_of_nonterm sym with
         | None -> rem
         | Some i ->
-          if different_levels (fst from,level) (where,i) then
+          if different_levels (from.notation_entry,level) (where,i) then
             (where,int_of_string i) :: rem
           else rem)
      | _ -> rem in
@@ -569,16 +579,16 @@ let make_act : type r. r target -> _ -> r gen_eval = function
   let env = (env.constrs, env.constrlists, env.binders, env.binderlists) in
   CAst.make ~loc @@ CNotation (None, notation, env)
 | ForPattern -> fun notation loc env ->
-  let env = (env.constrs, env.constrlists) in
+  let env = (env.constrs, env.constrlists, env.binders) in
   CAst.make ~loc @@ CPatNotation (None, notation, env, [])
 
 let extend_constr state forpat ng =
-  let custom,n,_ = ng.notgram_level in
+  let {notation_entry = custom; notation_level = _} as fromlev,_ = ng.notgram_level in
   let assoc = ng.notgram_assoc in
-  let (entry, level) = interp_constr_entry_key custom forpat n in
+  let (entry, level) = interp_constr_entry_key fromlev forpat in
   let fold (accu, state) pt =
-    let AnyTyRule r = make_ty_rule assoc (custom,n) forpat pt in
-    let pure_sublevels = pure_sublevels' assoc (custom,n) forpat level pt in
+    let AnyTyRule r = make_ty_rule assoc fromlev forpat pt in
+    let pure_sublevels = pure_sublevels' assoc fromlev forpat level pt in
     let isforpat = target_to_bool forpat in
     let needed_levels, state = register_empty_levels state isforpat pure_sublevels in
     let (pos,p4assoc,name,reinit), state = find_position state custom isforpat assoc level in
@@ -609,7 +619,7 @@ let extend_constr state forpat ng =
   in
   List.fold_left fold ([], state) ng.notgram_prods
 
-let constr_levels = GramState.field ()
+let constr_levels = GramState.field "constr_levels"
 
 let is_disjunctive_pattern_rule ng =
   String.is_sub "( _ | " (snd ng.notgram_notation) 0
@@ -618,7 +628,7 @@ let warn_disj_pattern_notation =
   let open Pp in
   let pp ng = str "Use of " ++ Notation.pr_notation ng.notgram_notation ++
               str " Notation is deprecated as it is inconsistent with pattern syntax." in
-  CWarnings.create ~name:"disj-pattern-notation" ~category:"notation" ~default:CWarnings.Disabled pp
+  CWarnings.create ~name:"disj-pattern-notation" ~category:CWarnings.CoreCategories.syntax ~default:CWarnings.Disabled pp
 
 let extend_constr_notation ng state =
   let levels = match GramState.get state constr_levels with
@@ -638,6 +648,6 @@ let extend_constr_notation ng state =
   (r @ r', state)
 
 let constr_grammar : one_notation_grammar grammar_command =
-  create_grammar_command "Notation" extend_constr_notation
+  create_grammar_command "Notation" { gext_fun = extend_constr_notation; gext_eq = (==) (* FIXME *) }
 
 let extend_constr_grammar ntn = extend_grammar_command constr_grammar ntn
